import numpy as np
from scipy.optimize import minimize
from joblib import Parallel, delayed
from ase import Atoms
#------------------------------------------------------------------------------------------
# POTENTIAL PARAMETERS AND CHARGES
parameters = {
    'charges': {'Ti': 2.196, 'O': -1.098},
    'buckingham': {
        ('Ti', 'Ti'): (31120.1, 0.1540, 5.25),
        ('O', 'O'): (11782.7, 0.2340, 30.22),
        ('O', 'Ti'): (16957.5, 0.1940, 12.59)
    },
    'lennard_jones': {
        ('Ti', 'Ti'): (1.0, 0.0),
        ('O', 'O'): (1.0, 0.0),
        ('O', 'Ti'): (1.0, 0.0)
    }
}
COULOMB_K = 14.39964546866782058  # eV·A
#------------------------------------------------------------------------------------------
def buckingham_coulomb_energy_forces(positions, symbols, parameters):
    n = len(positions)
    energy = 0.0
    forces = np.zeros_like(positions)
    charges = np.array([parameters['charges'][s] for s in symbols])
    for i in range(n):
        pos_i = positions[i]
        charge_i = charges[i]
        symbol_i = symbols[i]
        for j in range(i + 1, n):
            r_vec = pos_i - positions[j]
            r_sq = np.dot(r_vec, r_vec)
            r = np.sqrt(r_sq)
            inv_r = 1.0 / r
            r_hat = r_vec * inv_r
            
            # Precompute common powers of r
            inv_r2 = inv_r * inv_r
            inv_r6 = inv_r2 * inv_r2 * inv_r2
            inv_r7 = inv_r6 * inv_r
            inv_r12 = inv_r6 * inv_r6
            inv_r13 = inv_r12 * inv_r
            
            # Get pair parameters
            pair_key = (min(symbol_i, symbols[j]), max(symbol_i, symbols[j]))
            A, rho, C = parameters['buckingham'][pair_key]
            A_lj, _ = parameters['lennard_jones'][pair_key]
            
            # Coulomb interaction
            coulomb_energy = COULOMB_K * charge_i * charges[j] * inv_r
            coulomb_force = coulomb_energy * inv_r
            
            # Buckingham interaction
            exp_factor = np.exp(-r / rho)
            buck_energy = A * exp_factor - C * inv_r6
            buck_force = (A / rho) * exp_factor - 6 * C * inv_r7
            
            # Lennard-Jones repulsive interaction
            lj_energy = A_lj * inv_r12
            lj_force = 12.0 * A_lj * inv_r13
            
            # Sum contributions
            energy += coulomb_energy + buck_energy + lj_energy
            force_scalar = coulomb_force + buck_force + lj_force
            
            forces[i] += force_scalar * r_hat
            forces[j] -= force_scalar * r_hat
            
    return energy, forces
#------------------------------------------------------------------------------------------
def energy_forces_wrapper(x, symbols, parameters):
    positions = x.reshape(-1, 3)
    energy, forces = buckingham_coulomb_energy_forces(positions, symbols, parameters)
    return energy, -forces.reshape(-1)
#------------------------------------------------------------------------------------------
def optimize_TiO2(atoms, gtol=1e-6):
    symbols = atoms.get_chemical_symbols()
    x0 = atoms.positions.reshape(-1)
    result = minimize(
        fun=energy_forces_wrapper,
        x0=x0,
        args=(symbols, parameters),
        jac=True,
        method='BFGS',
        options={'gtol': gtol, 'disp': False, 'maxiter': 500}
    )
    opt_positions = result.x.reshape(-1, 3)
    energy_final, _ = buckingham_coulomb_energy_forces(
        opt_positions, symbols, parameters
    )
    atoms_opt = atoms.copy()
    atoms_opt.set_positions(opt_positions)
    atoms_opt.info['e'] = energy_final
    atoms_opt.info['i'] = atoms.info['i']
    return atoms_opt
#------------------------------------------------------------------------------------------
def optimize_TiO2_parallel(mol_list, n_jobs=1):
    return Parallel(n_jobs=n_jobs)(delayed(optimize_TiO2)(mol) for mol in mol_list)
#------------------------------------------------------------------------------------------
def example():
    import time
    from ase.io import read
    from aegon.libutils import writexyzs, rename, sort_by_energy
    clusters = read("stage1.xyz", index=":")
    print('PARALLEL OPTIMIZATION')
    rename(clusters, 'X', 3)
    for i in range(3):
        start_time = time.time()
        clusters = optimize_TiO2_parallel(clusters, n_jobs=10)
        #clusters = sort_by_energy(clusters, 1)
        end_time = time.time()
        print("Total time = %.2f" % (end_time-start_time))
    writexyzs(clusters, 'optimizados.xyz')
#example()



from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.core import Molecule
from pymatgen.symmetry.analyzer import PointGroupAnalyzer
#------------------------------------------------------------------------------------------
def molase2pymatgen(moleculease):
    if any(moleculease.pbc):
        structurepymatgen = AseAtomsAdaptor.get_structure(moleculease)
        return structurepymatgen
    else:
        symbols = moleculease.get_chemical_symbols()
        positions = moleculease.get_positions()
        moleculepymatgen=Molecule(symbols, positions)
        return moleculepymatgen
#------------------------------------------------------------------------------------------
def point_group(moleculease, tolerance=0.25, eigen_tolerance=1E-3, matrix_tolerance=0.1):
    molpmg=molase2pymatgen(moleculease)
    molx=PointGroupAnalyzer(molpmg, tolerance, eigen_tolerance, matrix_tolerance)
    pointgroup=molx.get_pointgroup()
    return pointgroup
#------------------------------------------------------------------------------------------
def symmetry_number(moleculease, tolerance=0.25, eigen_tolerance=1E-3, matrix_tolerance=0.1):
    molpmg=molase2pymatgen(moleculease)
    molx=PointGroupAnalyzer(molpmg, tolerance, eigen_tolerance, matrix_tolerance)
    symmetry_number=len(molx.get_symmetry_operations())
    return symmetry_number
#------------------------------------------------------------------------------------------
def inequivalent_finder(moleculease, tolerance=0.3, eigen_tolerance=0.008, matrix_tolerance=0.1):
    molpmg=molase2pymatgen(moleculease)
    #molpmg=AseAtomsAdaptor.get_structure(moleculease)
    molx=PointGroupAnalyzer(molpmg, tolerance, eigen_tolerance, matrix_tolerance)
    dictionary=molx.get_equivalent_atoms()
    dict=dictionary['eq_sets']
    inequiv=[ix for ix in dict.keys()]
    #for ix in dict.keys(): print(list(dict[ix]))
    #print('')
    inequiv.sort()
    return inequiv
#------------------------------------------------------------------------------------------



import time
import numpy as np
from joblib import Parallel, delayed
from scipy.optimize import minimize
from numba import jit
#------------------------------------------------------------------------------------------
epsilon=1.0
sigma = 3.0 / np.power(2.0, 1.0/6.0)
cutoff=18.0
method='L-BFGS-B'
gtol=1e-6
#------------------------------------------------------------------------------------------
@jit(nopython=True)
def lj_energy(positions):
    N = len(positions)
    energy = 0.0
    for i in range(N):
        for j in range(i + 1, N):
            rij = positions[i] - positions[j]
            r = np.linalg.norm(rij)
            if r < cutoff:
                sr6 = (sigma / r) ** 6
                energy += 4.0 * epsilon * sr6 * (sr6 - 1.0)
    return energy
#------------------------------------------------------------------------------------------
def lj_get_energy(atoms):
    positions = atoms.get_positions()
    return lj_energy(positions)
#------------------------------------------------------------------------------------------
@jit(nopython=True)
def lj_forces(positions):
    N = len(positions)
    forces = np.zeros_like(positions)
    for i in range(N):
        for j in range(i + 1, N):
            rij = positions[i] - positions[j]
            r = np.linalg.norm(rij)
            if r < cutoff:
                sr6 = (sigma / r) ** 6
                f_scalar = 24.0 * epsilon * sr6 * (2.0 * sr6 - 1.0) / r
                fij = f_scalar * (rij / r)
                forces[i] += fij
                forces[j] -= fij
    return forces
#------------------------------------------------------------------------------------------
def opt_lj(atoms):
    cluster = atoms.copy()
    x0 = cluster.get_positions()
    def objective(x):
        positions = x.reshape(-1, 3)
        return lj_energy(positions)
    def gradient(x):
        positions = x.reshape(-1, 3)
        forces = lj_forces(positions)
        return -forces.reshape(-1)
    result = minimize(
        fun=objective,
        x0=x0.reshape(-1),
        jac=gradient,
        method=method,
        options={'disp': False, 'gtol': gtol}
    )
    opt_positions = result.x.reshape(-1, 3)
    energy_scipy = result.fun
    cluster_opt = atoms.copy()
    cluster_opt.set_positions(opt_positions)
    cluster_opt.info['e'] = energy_scipy
    cluster_opt.info['c'] = 1 if result.success else 0
    return cluster_opt
#------------------------------------------------------------------------------------------
def opt_LJ_parallel(mol_list, n_jobs=-1):
    start_time = time.time()
    n1=len(mol_list)
    if not isinstance(mol_list, list): mol_list = [mol_list]
    results = Parallel(n_jobs=n_jobs)(delayed(opt_lj)(mol) for mol in mol_list)
    n2=len(results)
    end_time = time.time()
    print("Local OPT parallel at %.2f s [%d -> %d]" % (end_time-start_time, n1, n2))
    return results
#------------------------------------------------------------------------------------------
#if __name__ == "__main__":
#    from aegon.libutils import readxyzs
#    lj_clusters=readxyzs('Wales003to150.xyz')
#    for imol in lj_clusters:
#        n=len(imol)
#        energy_true= imol.info['e']
#        imol=opt_lj(imol)
#        energy_lj= imol.info['e']
#        print("#%s E_x = %13.8f E_True= %13.8f" %(str(n).zfill(3), energy_lj, energy_true))



import re
import os.path
import numpy as np
from ase import Atom, Atoms
from ase.data import chemical_symbols
#------------------------------------------------------------------------------------------
hartree2eV = 27.211386245981 #NIST
bohr2angstrom=0.529177210544 #NIST
eVtokcalpermol=23.060548012069496
hartree2kcalmol=627.5094738898777
#------------------------------------------------------------------------------------------
def get_termination_gaussian(pathfilename):
    if os.path.isfile(pathfilename):
        normal=0
        gaufile=open(pathfilename,'r')
        for line in gaufile:
            if "Normal termination" in line: normal=normal+1
        gaufile.close()
        return normal
    else:
        return -1
#------------------------------------------------------------------------------------------
def get_energy_gaussian(filename):
    enehartree=float(0.0)
    gaufile=open(filename,'r')
    for line in gaufile:
        if "SCF Done" in line:
            scf=line.split()
            enehartree=float(scf[4])
    gaufile.close()
    enekcalmol=enehartree*hartree2kcalmol
    #eneeV = enehartree * hartree2eV
    return enekcalmol
#------------------------------------------------------------------------------------------
def get_correction_zpe(filename):
    enehartree=float(0.0)
    gaufile=open(filename,'r')
    for line in gaufile:
        if "Zero-point correction=" in line: 
            zpc=line.split()
            enehartree=float(zpc[2])
    gaufile.close()
    enekcalmol=enehartree*hartree2kcalmol
    return enekcalmol
#------------------------------------------------------------------------------------------
def get_geometry_gaussian(pathfilename):
    nt=get_termination_gaussian(pathfilename)
    if nt==-1: return False
    energy=get_energy_gaussian(pathfilename)
    filename = os.path.basename(pathfilename)
    namein=filename.split('.')[0] 
    gaufile=open(pathfilename,'r')
    for line in gaufile:
        if line.strip() in ("Input orientation:", "Standard orientation:"):
            moleculeout = Atoms()
            moleculeout.info['c'] = nt
            moleculeout.info['e'] = energy
            moleculeout.info['i'] = namein
            for ii in range(4): line=gaufile.readline()
            line=gaufile.readline()
            while not line.startswith(" --------"):
                ls = line.split()
                if (len(ls) == 6 and ls[0].isdigit() and ls[1].isdigit() and ls[2].isdigit()):
                    numero_atomico=int(ls[1])
                    ss = chemical_symbols[numero_atomico]
                    xc,yc,zc = float(ls[3]), float(ls[4]), float(ls[5])
                    ai=Atom(symbol=ss, position=(xc, yc, zc))
                    moleculeout.append(ai)
                else:
                    break
                line=gaufile.readline()
                ls = line.split()
    gaufile.close()
    return moleculeout
#------------------------------------------------------------------------------------------
def get_traj_gaussian(pathfilename, force=False):
    nt=get_termination_gaussian(pathfilename)
    if nt==-1: return False
    filename=os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    start, end, ene, start_2, end_2 = [], [], [], [], []
    openold = open(pathfilename,"r")
    rline = openold.readlines()
    for i in range(len(rline)):
        if "Standard orientation:" in rline[i]:
            start.append(i+5)
            for j in range(i+5, len(rline)):
                if rline[j].strip().startswith("-"):
                    end.append(j - 1)
                    break
        if "Forces (Hartrees/Bohr)" in rline[i] and force:
            start_2.append(i+3)
            for j in range(i + 3, len(rline)):
                if rline[j].strip().startswith("-"):
                    end_2.append(j - 1)
                    break
        if "SCF Done" in rline[i]:
            eneline = rline[i].split()
            ene.append(eneline[4])      
    moleculeout=[]
    for i,iStart in enumerate(start[:-1]):
        enehartree=float(ene[i])
        eneeV = enehartree * hartree2eV
        singlemol = Atoms()
        singlemol.info['e'] = eneeV
        singlemol.info['c'] = nt
        singlemol.info['i'] = namein+'_'+str(i+1).zfill(3)
        for line in rline[start[i] : end[i]+1]:
            words = line.split() 
            numero_atomico = int(words[1])
            ss = chemical_symbols[numero_atomico]
            xc,yc,zc = float(words[3]), float(words[4]), float(words[5])
            ai=Atom(symbol=ss,position=(xc, yc, zc))
            singlemol.append(ai)
        if force:           
            forces_list_by_group = []
            for line in rline[start_2[i] : end_2[i]+1]:
                words = line.split()
                fx,fy,fz = float(words[2]), float(words[3]), float(words[4])
                fx=fx*hartree2eV/bohr2angstrom
                fy=fy*hartree2eV/bohr2angstrom
                fz=fz*hartree2eV/bohr2angstrom
                #IN ev/A
                forces_list_by_group.append([fx,fy,fz])
            singlemol.arrays['forces'] = np.array(forces_list_by_group)
        moleculeout.extend([singlemol])
    openold.close()
    return (moleculeout)
#------------------------------------------------------------------------------------------
def get_point_group_gaussian(filename):
    point_group = None
    with open(filename, 'r') as gaufile:
        for line in gaufile:
            if "Full point group" in line:
                parts = line.split()
                point_group = parts[-3]
    return point_group
#------------------------------------------------------------------------------------------
def get_state_spectroscopic_gaussian(filename):
    with open(filename, 'r') as f:
        for line in f:
            if "The electronic state is" in line:
                m = re.search(r"The electronic state is\s*([0-9]+)-(.+?)[\s\.]*$", line)
                if m:
                    multiplicity = m.group(1).strip()
                    symmetry = m.group(2).strip()
                    return f"{multiplicity}{symmetry}"
    return None
#------------------------------------------------------------------------------------------
def get_freqneg_gaussian(pathfilename):
    gaufile=open(pathfilename,'r')
    freq_negative=0
    freq_neg_list=[]
    for line in gaufile:
        if "Frequencies" in line:
            freq=line.split()[2:]
            for ifreq in freq:
                if float(ifreq) < 0.0:
                    freq_negative=freq_negative+1
                    freq_neg_list.append(float(ifreq))
    gaufile.close()
    freq_neg_list.sort()
    freq_sample= freq_neg_list[0] if (freq_negative > 0) else 0.0
    return freq_negative, freq_sample
#------------------------------------------------------------------------------------------
def get_freqlist(pathfilename):
    filename = os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    cfreq=0
    freq_list=[]
    grupo_atoms = {}
    gaufile=open(pathfilename,'r')
    for line in gaufile:
        if "Frequencies" in line:
            freq=line.split()[2:]
            nfreq=len(freq)
            freq_list = freq_list + [float(ifreq) for ifreq in freq]
            for ii in range(5): line=gaufile.readline()
            while len(line.split()) > 3:
                coords=line.split()
                ss=chemical_symbols[int(coords[1])]
                for jj, jfreq in enumerate(range(cfreq, nfreq+cfreq)):
                    xyz=coords[3*jj+2:3*jj+3+2]
                    dxc,dyc,dzc=float(xyz[0]),float(xyz[1]),float(xyz[2])
                    ai=Atom(symbol=ss, position=(dxc, dyc, dzc))
                    grupo_atoms.setdefault(jfreq, []).append(ai)
                line=gaufile.readline()
            cfreq=cfreq+nfreq
    gaufile.close()
    moleculeout = [Atoms(atomos) for atomos in grupo_atoms.values()]
    for i in range(len(moleculeout)):
        moleculeout[i].info['e'] = freq_list[i]
        moleculeout[i].info['i'] = namein+str(i+1).zfill(3)
    return moleculeout
#------------------------------------------------------------------------------------------
def operador_pm(moleculein, moleculedt, amplitud):
    moleculeout=moleculein.copy()
    moleculeout.set_positions(moleculeout.get_positions() + amplitud*moleculedt.get_positions())
    return moleculeout
#------------------------------------------------------------------------------------------
def displace_along_imaginary_mode(pathfilename, amplitud=0.4):
    mol0=get_geometry_gaussian(pathfilename)
    mold=get_freqlist(pathfilename)
    if mold[0].info['e'] < float(0.0):
        #SE TOMA LA MAS NEGATIVA
        molpm=operador_pm(mol0, mold[0], amplitud)
        return molpm
    else:
        return False
#from aegon.libcodegaussian import displace_along_imaginary_mode
#from aegon.libutils import writexyzs
#molx=displace_along_imaginary_mode('smi012.out')
#writexyzs([molx],'tmp.xyz')
#------------------------------------------------------------------------------------------
def make_a_input(singlemol, level='#WB97XD def2TZVP OPT SCF=(XQC) FREQ' , folder='./'):
    nameinp=singlemol.info['i']+'.inp'
    fh=open(folder+nameinp,"w")
    print("%NprocShared=13", file=fh)
    print("%MEM=16GB", file=fh)
    print("%s\n" %(level), file=fh)
    print("Comment: %s\n" %(singlemol.info['i']), file=fh)
    print("%d %d" %(singlemol.info['q'],singlemol.info['m']), file=fh)
    for iatom in singlemol:
        sym = iatom.symbol
        xc, yc, zc = iatom.position
        print ("%-2s %16.9f %16.9f %16.9f" % (sym,xc,yc,zc), file=fh)
    fh.write("\n")
    fh.close()
#------------------------------------------------------------------------------------------



import os
import time
import numpy as np
from multiprocessing import Process
from ase.data import atomic_masses
from aegon.libutils import readxyzs, writexyzs, centroid, sort_by_energy, prepare_folders, split_poscarlist, align_two
tolsij=0.95
tolene=0.10
#------------------------------------------------------------------------------------------
#Ultrafast Shape Recognition (USR) Algorithm
#------------------------------------------------------------------------------------------
def lastwo_central_moment(listxxx):
    xavg=np.mean(listxxx)
    res = listxxx - xavg
    ssq2 = np.mean(res**2) 
    ssq3 = np.mean(res**3) 
    return [ssq2,ssq3]
#------------------------------------------------------------------------------------------
def four_points(atoms):
    ####ctd: the molecular centroid
    ####cst: The closest atom to the molecular centroid (ctd)
    ####fct: The farthest atom from the molecular centroid (ctd)
    ####ftf: The farthest atom from the fct
    vecpos=[np.array(position) for position in atoms.get_positions()]
    ctd=centroid(atoms)
    r=[[np.linalg.norm(xyz - ctd), xyz] for xyz in vecpos]
    r.sort(key=lambda x: x[0])
    cst=r[0][1]
    fct=r[-1][1]
    s=[[np.linalg.norm(xyz - fct), xyz] for xyz in vecpos]
    s.sort(key=lambda x: x[0])
    ftf=s[-1][1]
    return ctd, cst, fct, ftf
#------------------------------------------------------------------------------------------
def USRMonoatom(moleculein):
    ctd,   cst,  fct,  ftf=four_points(moleculein)
    lctd, lcst, lfct, lftf= [], [], [], []
    for iatom in moleculein:
        xyz=np.array(iatom.position)
        ictd1=np.linalg.norm(xyz-ctd)
        icst1=np.linalg.norm(xyz-cst)
        ifct1=np.linalg.norm(xyz-fct)
        iftf1=np.linalg.norm(xyz-ftf)
        lctd.append(ictd1)
        lcst.append(icst1)
        lfct.append(ifct1)
        lftf.append(iftf1)
    a1=lastwo_central_moment(lctd)
    a2=lastwo_central_moment(lcst)
    a3=lastwo_central_moment(lfct)
    a4=lastwo_central_moment(lftf)
    standard_USR=a1+a2+a3+a4
    return standard_USR
#------------------------------------------------------------------------------------------
def USRMultiatom(moleculein):
    listm=moleculein.get_masses()
    mavg = np.mean(listm)
    ctd, cst, fct, ftf=four_points(moleculein)
    lctd,  lcst,  lfct,  lftf = [], [], [], []
    lctdm, lcstm, lfctm, lftfm = [], [], [], []
    for iatom in moleculein:
        xyz=np.array(iatom.position)
        ictd1=np.linalg.norm(xyz-ctd)
        icst1=np.linalg.norm(xyz-cst)
        ifct1=np.linalg.norm(xyz-fct)
        iftf1=np.linalg.norm(xyz-ftf)
        lctd.append(ictd1)
        lcst.append(icst1)
        lfct.append(ifct1)
        lftf.append(iftf1)
        mi=atomic_masses[iatom.number]
        ictdm1=ictd1*mi/mavg
        icstm1=icst1*mi/mavg
        ifctm1=ifct1*mi/mavg
        iftfm1=iftf1*mi/mavg
        lctdm.append(ictdm1)
        lcstm.append(icstm1)
        lfctm.append(ifctm1)
        lftfm.append(iftfm1)
    a1=lastwo_central_moment(lctd)
    a2=lastwo_central_moment(lcst)
    a3=lastwo_central_moment(lfct)
    a4=lastwo_central_moment(lftf)
    b1=lastwo_central_moment(lctdm)
    b2=lastwo_central_moment(lcstm)
    b3=lastwo_central_moment(lfctm)
    b4=lastwo_central_moment(lftfm)
    standard_USR=a1+a2+a3+a4
    m_weight_USR=b1+b2+b3+b4
    extra_descriptor_USR=standard_USR+m_weight_USR
    return extra_descriptor_USR
#------------------------------------------------------------------------------------------
def similarity(imol, jmol, mono=False):
    if mono:
        n_features=8
        vi=USRMonoatom(imol)
        vj=USRMonoatom(jmol)
    else:
        n_features=16
        vi=USRMultiatom(imol)
        vj=USRMultiatom(jmol)
    manhattan_distance=sum([np.absolute(a-b) for a,b in zip(vi,vj)])
    sij=1.0/(1.0+manhattan_distance/float(n_features))
    return sij
#------------------------------------------------------------------------------------------
def find_similar_elements(listmol, similarity_matrix, tols=tolsij, tole=tolene):
    similar_elements_indices = []
    num_elements = similarity_matrix.shape[0]
    debug=[]
    for i in range(num_elements):
        for j in range(i+1,num_elements):
            edf=np.abs(listmol[i].info['e'] - listmol[j].info['e'])
            if (similarity_matrix[i, j] >= tols) and (edf <= tole):
                similar_elements_indices.append(j)
                x1=listmol[j].copy()
                x2=listmol[i].copy()
                y1,y2=align_two(x1, x2)
                debug.extend([y1])
                debug.extend([y2])
    similar_elements_indices.sort()
    disimilars_atoms=[listmol[i] for i in range(num_elements) if i not in similar_elements_indices]
    if len(debug) > 0: writexyzs(debug, 'debug.xyz')
    return disimilars_atoms
#------------------------------------------------------------------------------------------
def disc_USR(listmol, tols=tolsij, tole=tolene, mono=False):
    num_molecules = len(listmol)
    if mono:
        n_features=8
        descriptors=[USRMonoatom(imol) for imol in listmol]
    else:
        n_features=16
        descriptors=[USRMultiatom(imol) for imol in listmol]
    #similar_elements_indices = []
    similarity_matrix = np.zeros((num_molecules, num_molecules))
    for i in range(num_molecules):
        vi=descriptors[i]
        for j in range(i, num_molecules):
            vj=descriptors[j]
            manhattan_distance=sum([np.absolute(a-b) for a,b in zip(vi,vj)])
            similarity=1.0/(1.0+manhattan_distance/float(n_features))
            similarity_matrix[i, j] = similarity
            similarity_matrix[j, i] = similarity
    disimilars_atoms=find_similar_elements(listmol, similarity_matrix, tols, tole)
    return disimilars_atoms
#------------------------------------------------------------------------------------------
def comparator_usr_serial(atoms0, tols=tolsij, tole=tolene, mono=False):
    start = time.time()
    ni=len(atoms0)
    atoms1=disc_USR(atoms0, tols, tole, mono)
    nf=len(atoms1)
    end = time.time()
    print('USR comparison (-serial-) at %5.2f s [%d -> %d]' %(end - start, ni, nf))
    return atoms1
#------------------------------------------------------------------------------------------
def make_comparator_usr(ifolder, tols=tolsij, tole=tolene, mono=False):
    atoms0=readxyzs(ifolder+'/'+ifolder+'.xyz')
    atoms1=disc_USR(atoms0, tols, tole, mono)
    writexyzs(atoms1,ifolder+'/'+ifolder+'_disc.xyz')
#------------------------------------------------------------------------------------------
def comparator_usr_parallel(poscarlist, nproc, base_name, tols=tolsij, tole=tolene, mono=False):
    poscarlist=sort_by_energy(poscarlist, 1)
    ni=len(poscarlist)
    if ni < 4*nproc:
        moleculeout=comparator_usr_serial(poscarlist, tols, tole, mono)
        moleculeout=sort_by_energy(moleculeout, 1)
        return moleculeout
    start = time.time()
    folderlist=prepare_folders(poscarlist, nproc, base_name)
    poscar_split_list=split_poscarlist(poscarlist, nproc)
    procs = []
    dicc_term = {iposcar.info['i']: iposcar.info['c'] for iposcar in poscarlist}
    for ifolder, iposcars in zip(folderlist, poscar_split_list):
        writexyzs(iposcars, ifolder+'/'+ifolder+'.xyz')
        proc = Process(target=make_comparator_usr, args=(ifolder, tols, tole, mono))
        procs.append(proc)
        proc.start()
    for proc in procs:
        proc.join()
    center_mols, border_mols = [], []
    for ifolder in folderlist:
        molx=readxyzs(ifolder+'/'+ifolder+'_disc.xyz')
        for imol in molx: imol.info['c']=dicc_term[imol.info['i']]
        molx=sort_by_energy(molx, 1)
        emin=molx[0].info['e']
        emax=molx[-1].info['e']
        for imol in molx:
            delw=imol.info['e'] - emin
            deup=emax - imol.info['e']
            if ( delw < tole ) or ( deup < tole ):
                border_mols.extend([imol])
            else:
                center_mols.extend([imol])
    border_mols=disc_USR(border_mols, tols, tole, mono)
    moleculeout=center_mols+border_mols
    moleculeout=sort_by_energy(moleculeout, 1)
    os.system('rm -rf %sproc[0-9][0-9][0-9]' %(base_name))
    nf=len(moleculeout)
    end = time.time()
    print('USR comparison (parallel) at %5.2f s [%d -> %d]' %(end - start, ni, nf))
    return moleculeout
#------------------------------------------------------------------------------------------
def make_similarity_matrix_compare(moleculein, moleculeref, mono=False):
    if mono:
        n_features=8
        descriptors1=[USRMonoatom(imol) for imol in moleculein]
        descriptors2=[USRMonoatom(imol) for imol in moleculeref]
    else:
        n_features=16
        descriptors1=[USRMultiatom(imol) for imol in moleculein]
        descriptors2=[USRMultiatom(imol) for imol in moleculeref]
    total_molecules1=len(moleculein)
    total_molecules2=len(moleculeref)
    similarity_matrix=np.zeros(shape=(total_molecules1, total_molecules2),dtype=float)
    for i in range(total_molecules1):
        vi=descriptors1[i]
        for j in range(total_molecules2):
            vj=descriptors2[j]
            manhattan_distance=sum([np.absolute(a-b) for a,b in zip(vi,vj)])
            similarity=1.0/(1.0+manhattan_distance/float(n_features))
            similarity_matrix[i, j] = similarity
    return similarity_matrix
#------------------------------------------------------------------------------------------
def molin_sim_molref(moleculein, moleculeref, tols=tolsij, tole=tolene, mono=False):
    start = time.time()
    ni=len(moleculein)
    matrixs=make_similarity_matrix_compare(moleculein, moleculeref, mono)
    similares=[]
    for i,imol in enumerate(moleculein):
        for j, jmol in enumerate(moleculeref):
            sij=matrixs[i][j]
            edf=np.abs(imol.info['e']-jmol.info['e'])
            if (sij >= tols) and (edf <= tole):
                similares.append(i)
                break
    moleculeout=[imol for i, imol in enumerate(moleculein) if i not in similares] 
    nf=len(moleculeout)
    end = time.time()
    print('USR comparison (-serial-) at %5.2f s [%d -> %d]' %(end - start, ni, nf))
    return moleculeout
#------------------------------------------------------------------------------------------



import os
import numpy as np
from ase import Atoms
from ase.data import covalent_radii
from aegon.libutils import align
#------------------------------------------------------------------------------------------
def tag(poscarlist):
    if not isinstance(poscarlist, list):
        poscarlist = [poscarlist]
    moleculeout = []
    for atoms in poscarlist:
        force = hasattr(atoms, "forces")
        numbers = list(set(atoms.get_atomic_numbers()))
        tagdict = {zi: itag for itag, zi in enumerate(sorted(numbers))}
        enumerate_molecule = atoms.copy()
        for iatom in enumerate_molecule:
            iatom.tag = tagdict[iatom.number]
        if force:
            enumerate_molecule.arrays['forces'] = atoms.arrays['forces']
        moleculeout.append(enumerate_molecule)
    return moleculeout
#------------------------------------------------------------------------------------------
def order_and_tag(poscarlist):
    if not isinstance(poscarlist, list):
        poscarlist = [poscarlist]
    outlist = []
    for atoms in poscarlist:
        numbers = atoms.get_atomic_numbers()
        symbols = atoms.get_chemical_symbols()
        positions = atoms.get_positions()
        forces = atoms.arrays['forces'] if 'forces' in atoms.arrays else None
        idx = np.argsort(numbers)
        sorted_numbers = [numbers[i] for i in idx]
        sorted_symbols = [symbols[i] for i in idx]
        sorted_positions = [positions[i] for i in idx]
        if forces is not None:
            sorted_forces = [forces[i] for i in idx]
        sorted_atoms = Atoms(symbols=sorted_symbols,
                             positions=sorted_positions,
                             cell=atoms.cell,
                             pbc=atoms.pbc)
        unique_numbers = sorted(set(sorted_numbers))
        tagdict = {zi: itag for itag, zi in enumerate(unique_numbers)}
        sorted_atoms.set_tags([tagdict[zi] for zi in sorted_numbers])
        if forces is not None:
            sorted_atoms.set_array("forces", np.array(sorted_forces))
        outlist.append(sorted_atoms)
    return outlist
# ------------------------------------------------------------------------------------------
def make_matrix(mol, latsp=5.0):
    mol_centered = mol.copy()
    mol_centered.translate(-mol_centered.get_center_of_mass())
    pos = mol_centered.get_positions()
    radii = [covalent_radii[z] for z in mol_centered.get_atomic_numbers()]
    extent = np.max(np.abs(pos) + np.array(radii)[:, None], axis=0) * 2.0 + latsp
    return np.diag(extent)
# ------------------------------------------------------------------------------------------
def molecule2poscar(atomslist, latsp=5.0, do_align=False):
    if not isinstance(atomslist, list):
        atomslist = [atomslist]
    outlist = []
    for mol in atomslist:
        mol_copy = mol.copy()
        if do_align: align(mol_copy)
        mol_copy.translate(-mol_copy.get_center_of_mass())
        matrix = make_matrix(mol_copy, latsp)
        shift = 0.5 * np.diag(matrix)
        mol_copy.translate(shift)
        mol_copy.set_cell(matrix)
        mol_copy.set_pbc([True, True, True])
        outlist.append(mol_copy)
    return order_and_tag(outlist)
# ------------------------------------------------------------------------------------------
def readposcars(filename):
    if not os.path.isfile(filename):
        print("The file %s does not exist." %(filename))
        exit()
    contcarfile=open(filename,'r')
    poscarout=[]
    for line in contcarfile:
        header=line.split()
        name=header[0]
        energy=float(header[1]) if len(header)>1 else 0.0
        line=contcarfile.readline()
        if str(line.split()[0])=='0.00000000E+00': break
        #-----------------------------------
        scale=float(line.split()[0])
        line=contcarfile.readline()
        a1x, a1y, a1z=map(float,line.split())
        line=contcarfile.readline()
        a2x, a2y, a2z=map(float,line.split())
        line=contcarfile.readline()
        a3x, a3y, a3z=map(float,line.split())
        lattice_vectors=np.array([[a1x, a1y, a1z],[a2x, a2y, a2z],[a3x, a3y, a3z]])*scale
        #-----------------------------------
        line=contcarfile.readline()
        elements=line.split()
        line=contcarfile.readline()
        ocupnumchar=line.split()
        ocupnuminte=list(map(int, ocupnumchar))
        #-----------------------------------
        natom=sum(ocupnuminte)
        liste,kk=[],0
        for ii in ocupnuminte:
            for jj in range(ii):
                liste.append(elements[kk])
            kk=kk+1
        #-----------------------------------
        line=contcarfile.readline()
        sd=0
        if 'Selective dynamics' in line:
            line=contcarfile.readline()
            sd=1
        symbols=[]
        positions=[]
        if 'Direct' in line:
            for iatom in range(natom):
                line=contcarfile.readline()
                vecxyz=line.split()
                s=liste[iatom]
                symbols.append(s)
                xd=float(vecxyz[0])
                yd=float(vecxyz[1])
                zd=float(vecxyz[2])
                direct_coords = np.array([xd, yd, zd])
                cart_coords = np.dot(direct_coords, lattice_vectors)
                positions.append(cart_coords)
        if 'Cartesian' in line:
            for iatom in range(natom):
                s=liste[iatom]
                symbols.append(s)
                line=contcarfile.readline()
                cart_coords =np.array(line.split()[0:3])
                positions.append(cart_coords)
        positions=np.array(positions)
        iposcar = Atoms(symbols=symbols, positions=positions, cell=lattice_vectors, pbc=True)
        iposcar.info['i']=name
        iposcar.info['e']=energy
        poscarout.extend([iposcar])
    contcarfile.close()
    return poscarout
#------------------------------------------------------------------------------------------
def writeposcars(poscarlist, file, opt='D'):
    with open(file, "w") as f:
        for atoms in poscarlist:
            print("%s %12.8f" % (atoms.info['i'], atoms.info['e']), file=f)
            print('1.0', file=f)
            matrix = atoms.cell.array
            for vec in matrix:
                print("%20.16f %20.16f %20.16f" % tuple(vec), file=f)
            element_count = {}
            for sym in atoms.symbols:
                element_count[sym] = element_count.get(sym, 0) + 1
            print(' '.join(element_count.keys()), file=f)
            print(' '.join(str(v) for v in element_count.values()), file=f)
            if opt == 'C':
                print('Cartesian', file=f)
                count = {}
                for atom in atoms:
                    sym = atom.symbol
                    count[sym] = count.get(sym, 0) + 1
                    x, y, z = atom.position
                    print("%20.16f %20.16f %20.16f   !%s%d" %
                          (x, y, z, sym, count[sym]), file=f)
            elif opt == 'D':
                print('Direct', file=f)
                count = {}
                for atom in atoms:
                    sym = atom.symbol
                    count[sym] = count.get(sym, 0) + 1
                    frac = np.linalg.solve(matrix.T, atom.position)
                    print("%20.16f %20.16f %20.16f   !%s%d" %(frac[0], frac[1], frac[2], sym, count[sym]), file=f)
#------------------------------------------------------------------------------------------
def conventional(atoms, tol=0.0):
    dup_tol = 1e-5
    lattice = atoms.cell.array
    frac_coords = atoms.get_scaled_positions(wrap=False)
    symbols = atoms.get_chemical_symbols()
    new_symbols = []
    new_positions = []
    shifts = np.array(np.meshgrid([-1, 0, 1],[-1, 0, 1],[-1, 0, 1])).T.reshape(-1, 3)
    for icoord, sym in zip(frac_coords, symbols):
        for shift in shifts:
            fcoord = icoord + shift
            if np.all((fcoord >= -tol) & (fcoord <= 1.0 + tol)):
                pos = fcoord @ lattice
                if any(np.allclose(pos, p, atol=dup_tol) and s == sym
                       for p, s in zip(new_positions, new_symbols)):
                    continue
                new_symbols.append(sym)
                new_positions.append(pos)
    poscarout = Atoms(symbols=new_symbols, positions=new_positions, cell=lattice, pbc=True)
    return poscarout
#------------------------------------------------------------------------------------------



import os.path
from ase.data import atomic_masses, atomic_numbers
#------------------------------------------------------------------------------------------
def sort_composition(composition):
    s=[[x[0],x[1], atomic_masses[atomic_numbers[x[0]]]] for x in composition]
    t = sorted(s, key=lambda x: float(x[2]), reverse=True)
    newcomposition=([[x[0],x[1]] for x in t])
    return newcomposition
#------------------------------------------------------------------------------------------
def get_inatoms(composition):
    newcomposition=sort_composition(composition)                                
    inatoms=[]
    for xxx in newcomposition:
        for jj in range(xxx[1]):
            inatoms.append(xxx[0])
    return inatoms
#------------------------------------------------------------------------------------------
def clustername(composition):
    newcomposition=sort_composition(composition)
    chainname=' '.join([item[0]+str(item[1]) for item in newcomposition])
    return chainname
#------------------------------------------------------------------------------------------
def get_elements(composition):
    newcomposition=sort_composition(composition)
    atoms=([x[0] for x in newcomposition])
    return atoms
#------------------------------------------------------------------------------------------
#df = read_main_input(inputfile)
#cf=df.get_comp('composition').comp
def read_composition(filename, key):
    start_marker = f"---{key.upper()}---"
    end_marker = start_marker
    composition = []
    recording = False
    with open(filename, 'r') as file:
        for line in file:
            stripped = line.strip()
            if stripped == start_marker:
                recording = not recording
                continue
            if recording and stripped:
                parts = stripped.split()
                if len(parts) == 2:
                    element, count = parts[0], int(parts[1])
                    composition.append([element, count])
                else:
                    raise ValueError(f"Invalid line format in block '{key}': '{line.strip()}'")
    class Konpozition:
        def __init__(self, xcomposition):
            self.comp    = sort_composition(xcomposition)
            self.atoms   = get_inatoms(self.comp)
            self.name    = clustername(self.comp)
            self.elements= get_elements(self.comp)
    classcomp=Konpozition(composition)
    return classcomp
#------------------------------------------------------------------------------------------
#PENDIENTE DE ELIMINAR
def read_block_from_file_antiguo(filepath, id):
    bilfile=open(filepath,"r")
    chainchar='---'+id.upper()+'---'
    printer=0
    data_block=[]
    for line in bilfile:
         lin = line.lstrip()
         if lin.startswith(chainchar): printer=1+printer
         if printer == 1 and not lin.startswith(chainchar): data_block.append(line)
    bilfile.close()
    return data_block 
#------------------------------------------------------------------------------------------
def read_block_from_file(filepath, id):
    block = []
    inside = False
    chainchar='---'+id.upper()+'---'
    with open(filepath, "r") as f:
        for line in f:
            if line.strip() == chainchar:
                inside = not inside
                continue
            if inside:
                block.append(line.rstrip())
    #block="\n".join(block)
    #return block.splitlines()
    return block
#------------------------------------------------------------------------------------------
def get_value_from_file(filepath, key, dtype=str, default=None):
    """
    Reads the value associated with 'key' from a parameter file, ignoring comments.
    
    Parameters:
        filepath (str): Path to the input file.
        key (str): Name of the parameter to search for.
        dtype (type): Type of the value to return (int, float, str).
        default: Default value if key is not found.
    
    Returns:
        The parsed value of the specified type, or the default if not found.
    """
    if not os.path.isfile(filepath):
        return default
    with open(filepath, 'r') as f:
        for line in f:
            line = line.split('#')[0].strip()
            if not line:
                continue
            parts = line.split()
            if len(parts) >= 2 and parts[0].lower() == key.lower():
                value_str = parts[1]
                try:
                    return dtype(value_str)
                except ValueError:
                    print(f"Warning: Could not convert '{value_str}' to {dtype.__name__}. Returning default.")
                    return default
    return default
#------------------------------------------------------------------------------------------
def get_list_from_file(filepath, key, datatype, default=None):
    """
    Reads a list of values (int, float, or str) from a given key in an input file.
    Parameters:
        filepath (str): Path to the input file.
        key (str): Name of the variable to extract.
        datatype (type): Type to which each item should be cast (int, float, str).
        default (list): Default list to return if the key is not found.
    Returns:
        list: List of values of specified type.
    """
    if default is None:
        default = []
    if not os.path.isfile(filepath):
        return default
    with open(filepath, 'r') as f:
        for line in f:
            line = line.split('#')[0].strip()
            if not line:
                continue
            parts = line.split()
            if parts and parts[0].lower() == key.lower():
                try:
                    return [datatype(x) for x in parts[1:]]
                except ValueError:
                    print(f"Warning: Could not convert values in key '{key}' to {datatype.__name__}.")
                    return default
    return default
#------------------------------------------------------------------------------------------
class read_main_input:
    def __init__(self, filename):
        self.filename = filename
    def get_int(self, key, default):
        return get_value_from_file(self.filename, key, int, default)
    def get_float(self, key, default):
        return get_value_from_file(self.filename, key, float, default)
    def get_str(self, key, default):
        return get_value_from_file(self.filename, key, str, default)
    def get_int_list(self, key, default):
        return get_list_from_file(self.filename, key, int, default)
    def get_float_list(self, key, default):
        return get_list_from_file(self.filename, key, float, default)
    def get_str_list(self, key, default):
        return get_list_from_file(self.filename, key, str, default)
    def get_block(self, key):
        return read_block_from_file(self.filename, key)
    def get_comp(self, key, index=0):
        return read_composition(self.filename, key)
#------------------------------------------------------------------------------------------



import numpy as np
from importlib import resources
from ase import Atoms

# ============================================================
# Internal utilities
# ============================================================

def _load_npz(filename):
    """
    Load an .npz file from aegon/data/.
    """
    try:
        with resources.as_file(resources.files("aegon").joinpath(f"data/{filename}")) as path:
            return np.load(path, allow_pickle=True)
    except FileNotFoundError:
        raise ValueError(f"Data file not found: data/{filename}")


def _atoms_from_entry(data, N, symbol):
    """
    Build an ASE Atoms object from a database entry.
    """
    atoms = Atoms(symbols=[symbol] * N, positions=data["positions"])
    atoms.info["energy"] = data["energy"]
    atoms.info["label"] = data["label"]
    return atoms

# ============================================================
# ClusterFactory Core
# ============================================================

class ClusterFactory:

    # Cached databases to avoid reloading from disk
    _LJ_DATA = None
    _SC_DATA = {}   # key: symbol → db

    # Supported models (LJ and SC for now)
    _SUPPORTED_MODELS = {"LJ", "SC"}

    # ---------------------- Public API ----------------------

    @classmethod
    def get(cls, N, model, element=None):
        """
        Unified entry point for obtaining a cluster.

        Parameters
        ----------
        N : int
            Number of atoms.
        model : str
            "LJ" or "SC"
        element : str, optional
            Atomic symbol (required for SC).
        """
        model = model.upper()

        if model not in cls._SUPPORTED_MODELS:
            raise ValueError(f"Unsupported model '{model}'. Supported models: {cls._SUPPORTED_MODELS}")

        if model == "LJ":
            return cls._get_lj(N, element or "Mo")  # default LJ element is Mo

        elif model == "SC":
            if element is None:
                raise ValueError("Element must be specified for SC clusters.")
            return cls._get_sc(N, element)

    @classmethod
    def list_available(cls, model, element=None):
        """
        List available cluster sizes for a given model.

        Parameters
        ----------
        model : str
            "LJ" or "SC"
        element : str, optional
            Required for SC.
        """
        model = model.upper()

        if model not in cls._SUPPORTED_MODELS:
            raise ValueError(f"Unsupported model '{model}'.")

        if model == "LJ":
            return cls._list_available_lj()

        elif model == "SC":
            if element is None:
                raise ValueError("Element must be specified for SC clusters.")
            return cls._list_available_sc(element)

    # ---------------------- LJ subsystem ----------------------

    @classmethod
    def _load_lj_database(cls):
        """
        Load the LJ database once.
        """
        if cls._LJ_DATA is None:
            cls._LJ_DATA = _load_npz("libdata_lj.npz")
        return cls._LJ_DATA

    @classmethod
    def _get_lj(cls, N, symbol):
        db = cls._load_lj_database()
        key = f"LJ{N:03d}"

        if key not in db:
            raise ValueError(f"No LJ cluster available for N = {N}")

        data = db[key].item()
        return _atoms_from_entry(data, N, symbol)

    @classmethod
    def _list_available_lj(cls):
        db = cls._load_lj_database()
        return sorted(int(k[2:]) for k in db.keys())

    # ---------------------- SC subsystem ----------------------

    @classmethod
    def _load_sc_database(cls, element):
        element = element.capitalize()

        if element not in cls._SC_DATA:
            filename = f"SC_{element}_clusters_data.npz"
            cls._SC_DATA[element] = _load_npz(filename)

        return cls._SC_DATA[element]

    @classmethod
    def _get_sc(cls, N, element):
        element = element.capitalize()
        db = cls._load_sc_database(element)

        key = f"SC{N:03d}"
        if key not in db:
            raise ValueError(f"No SC cluster available for element {element} with N = {N}")

        data = db[key].item()
        return _atoms_from_entry(data, N, element)

    @classmethod
    def _list_available_sc(cls, element):
        element = element.capitalize()
        db = cls._load_sc_database(element)
        return sorted(int(k[2:]) for k in db.keys())



import random
import numpy as np
from ase import Atom, Atoms
from ase.data import atomic_numbers, covalent_radii
from aegon.libutils import distance_atom2atoms, distance_interatom, connected_graph, scale_coords, distance_atoms2atoms, merge_atoms, rand_unit_vector
#------------------------------------------------------------------------------------------
ratio_radii_min =0.95
ratio_radii_max =1.15
pvalue=float(ratio_radii_min+ratio_radii_max)/2.0
cte=(4.0*np.pi*np.sqrt(2.0))
reactivelist=['H', 'Li']
#------------------------------------------------------------------------------------------
def min_step(inatoms):
    min_covalent_radii=min([covalent_radii[atomic_numbers[si]] for si in inatoms])
    return min_covalent_radii
#------------------------------------------------------------------------------------------
def make_molecule_3Da(inatoms):
    atom_step=min_step(inatoms)
    onion_radii_factor=float(0.5)
    radius_of_container=float(0.5+(((len(inatoms)*3)/cte)**(1.0/3.0)))
    radius_of_container=onion_radii_factor*radius_of_container*float(1.1)
    atoms=list(inatoms)
    random.shuffle(atoms)
    si=atoms.pop()
    xi, yi, zi=random.uniform(-1, 1), random.uniform(-1, 1), random.uniform(-1, 1)
    mm3da=Atoms(symbols=[si], positions=[(xi, yi, zi)])
    while atoms:
        random.shuffle(atoms)
        si=atoms.pop()
        inside, test=100, 0
        while inside > 0:
            RR=100.0
            while RR > radius_of_container: 
                xx=random.uniform(-radius_of_container, radius_of_container)
                yy=random.uniform(-radius_of_container, radius_of_container)
                zz=random.uniform(-radius_of_container, radius_of_container)
                RR=np.sqrt((xx*xx)+(yy*yy)+(zz*zz))
            tryatom=Atom(symbol=si, position=(xx, yy, zz))
            dmin=distance_atom2atoms(tryatom,mm3da)
            if (dmin >= ratio_radii_min) and (dmin <= ratio_radii_max):
                inside=0
                mm3da.append(tryatom)
            if test > 5000:
                radius_of_container=radius_of_container + atom_step
                test=0
            test+=1
    mm3da.info['e'] = 0.0
    mm3da.info['i'] = 'make_molecule_3Da'
    return mm3da
#------------------------------------------------------------------------------------------
def make_molecule_3Db(inatoms):
    atom_step=min_step(inatoms)
    onion_radii_factor=float(1.5)
    radius_of_container=float(0.5+(((len(inatoms)*3)/cte)**(1.0/3.0)))
    radius_of_container=onion_radii_factor*radius_of_container*float(1.1)
    atoms=list(inatoms)
    random.shuffle(atoms)
    si=atoms.pop()
    xi, yi, zi=random.uniform(-1, 1), random.uniform(-1, 1), random.uniform(-1, 1)
    mm3db=Atoms(symbols=[si], positions=[(xi, yi, zi)])
    while atoms:
        random.shuffle(atoms)
        si=atoms.pop() 
        inside, test=100, 0
        while inside > 0:
            RR=100.0
            while RR > radius_of_container: 
                xx=random.uniform(-radius_of_container, radius_of_container)
                yy=random.uniform(-radius_of_container, radius_of_container)
                zz=random.uniform(-radius_of_container, radius_of_container)
                RR=np.sqrt((xx*xx)+(yy*yy)+(zz*zz))
            tryatom=Atom(symbol=si, position=(xx, yy, zz))
            dmin=distance_atom2atoms(tryatom, mm3db)
            if (dmin >= ratio_radii_min) and (dmin <= ratio_radii_max):
                inside=0
                mm3db.append(tryatom)
            if test > 8000:
                radius_of_container=radius_of_container + atom_step
                test=0
            test+=1
    mm3db.info['e'] = 0.0
    mm3db.info['i'] = 'make_molecule_3Db'
    return mm3db
#------------------------------------------------------------------------------------------
def make_molecule_2Da(inatoms):
    atom_step=min_step(inatoms)
    onion_radii_factor=float(1.25)
    radius_of_container=float(0.5+(((len(inatoms)*3)/cte)**(1.0/3.0)))
    radius_of_container= onion_radii_factor*radius_of_container*float(1.2)
    atoms=list(inatoms)
    random.shuffle(atoms)
    si=atoms.pop()
    xi, yi, zi=random.uniform(-0.25, 0.25), random.uniform(-0.25, 0.25), 0.0
    mm2da=Atoms(symbols=[si], positions=[(xi, yi, zi)])
    while atoms:
        random.shuffle(atoms)
        si=atoms.pop() 
        inside, test=100, 0
        while inside > 0:
            RR=100.0
            while RR > radius_of_container: 
                xx=random.uniform(-radius_of_container, radius_of_container)
                yy=random.uniform(-radius_of_container, radius_of_container)
                RR=np.sqrt((xx*xx)+(yy*yy))
            tryatom=Atom(symbol=si, position=(xx, yy, 0.0))
            dmin=distance_atom2atoms(tryatom, mm2da)
            if (dmin >= ratio_radii_min) and (dmin <= ratio_radii_max):
                inside=0
                mm2da.append(tryatom)
            if test > 7000:
                radius_of_container=radius_of_container + atom_step
                test=0
            test+=1
    mm2da.info['e'] = 0.0
    mm2da.info['i'] = 'make_molecule_2Da'
    return mm2da
#------------------------------------------------------------------------------------------
def make_molecule_sphere(inatoms):
    atom_step=min_step(inatoms)
    onion_radii_factor=float(0.70)
    radius_of_container=float(0.5+(((len(inatoms)*3)/cte)**(1.0/3.0)))
    radius_of_container=onion_radii_factor*radius_of_container*float(1.2)
    atoms=list(inatoms)
    random.shuffle(atoms)
    si=atoms.pop()
    phi=float(random.uniform(0.0, 2*np.pi))
    costheta=float(random.uniform(-1.0, 1.0))
    theta = np.arccos(costheta)
    x = radius_of_container * np.sin(theta) * np.cos(phi)
    y = radius_of_container * np.sin(theta) * np.sin(phi)
    z = radius_of_container * np.cos(theta)
    mmsp=Atoms(symbols=[si], positions=[(float(x), float(y), float(z))])    
    while atoms:
        random.shuffle(atoms)
        si=atoms.pop() 
        inside, test=100, 0
        while inside > 0:
            phi=float(random.uniform(0.0, 2*np.pi))
            costheta=float(random.uniform(-1.0, 1.0))
            theta = np.arccos(costheta)
            xx = radius_of_container * np.sin(theta) * np.cos(phi)
            yy = radius_of_container * np.sin(theta) * np.sin(phi)
            zz = radius_of_container * np.cos(theta)
            tryatom=Atom(symbol=si, position=(float(xx), float(yy), float(zz)))
            dmin=distance_atom2atoms(tryatom, mmsp)
            if (dmin >= ratio_radii_min) and (dmin <= ratio_radii_max):
                inside=0
                mmsp.append(tryatom)
            if test > 5000:
                test=0
                radius_of_container=radius_of_container + atom_step
            test+=1
    mmsp.info['e'] = 0.0
    mmsp.info['i'] = 'make_molecule_sphere'
    return mmsp
#------------------------------------------------------------------------------------------
def make_molecule_wire(inatoms):
    atoms=list(inatoms)
    random.shuffle(atoms)
    si=atoms.pop()
    xxx=float(0.0)
    mm2da=Atoms(symbols=[si], positions=[(float(0.0), float(0.0), float(0.0))])
    last=[si]
    while atoms:
        random.shuffle(atoms)
        si=atoms.pop()
        a1=covalent_radii[atomic_numbers[last[-1]]]
        a2=covalent_radii[atomic_numbers[si]]
        xxx=float(a1+a2)*pvalue + xxx
        tryatom=Atom(symbol=si, position=(xxx, float(0.0), float(0.0)))
        mm2da.append(tryatom)
        last.append(si)
    mm2da.info['e'] = 0.0
    mm2da.info['i'] = 'make_molecule_wire'
    return mm2da
#------------------------------------------------------------------------------------------
def make_molecule_circle(inatoms):
    atoms=list(inatoms)
    random.shuffle(atoms)
    radiomax=sum([covalent_radii[atomic_numbers[ii]] for ii in atoms])/(np.pi)
    si=atoms.pop()
    mcircle=Atoms(symbols=[si], positions=[(radiomax, 0.0, 0.0)])
    last=[si]
    angle=0.0
    while atoms:
        si=atoms.pop()
        a1=covalent_radii[atomic_numbers[last[-1]]]        
        a2=covalent_radii[atomic_numbers[si]]
        S=a1+a2
        last.append(si)
        theta=S/radiomax
        angle=angle+theta
        xx=radiomax*np.cos(angle)
        yy=radiomax*np.sin(angle)
        ai=Atom(symbol=si, position=(float(xx), float(yy), 0.0))
        mcircle.append(ai)
    rmin=distance_interatom(mcircle)
    factor=(ratio_radii_min+ratio_radii_max)/(2.0*rmin)
    mcircle=scale_coords(mcircle, factor)
    mcircle.info['e'] = 0.0
    mcircle.info['i'] = 'make_molecule_circle'
    return mcircle
#------------------------------------------------------------------------------------------
def make_molecule_two_rings(inatoms):
    listsh=list(inatoms)
    random.shuffle(listsh)
    circ1=int(np.ceil(len(listsh)/2.0))
    atoms1=list(listsh[0:circ1])
    atoms2=list(listsh[(circ1):])
    circulo1=make_molecule_circle(atoms1)
    circulo2=make_molecule_circle(atoms2)
    zzz, inside=0.05, 100
    while inside > 50:
        tmp=circulo2.copy()
        vz=np.array([0.0, 0.0, zzz])
        tmp.translate(-vz)
        dmin=distance_atoms2atoms(tmp, circulo1)
        if (dmin > ratio_radii_min):
            inside=0
            two_rings=merge_atoms([circulo1,tmp])            
        else: zzz = zzz + 0.01
    two_rings.info['e'] = 0.0
    two_rings.info['i'] = 'make_molecule_two_rings'
    return two_rings
#------------------------------------------------------------------------------------------
def make_molecule_helix(inatoms):
    atoms0=list(inatoms)
    len0=len(atoms0)
    random.shuffle(atoms0)
    #----------------------------------
    circ1=random.randint(2, 6) if len0 > 6 else random.randint(1, len0)
    atoms1=list(atoms0[0:circ1])
    radiomax=sum([covalent_radii[atomic_numbers[ii]] for ii in atoms1])*pvalue/(np.pi)
    #----------------------------------
    ref0=covalent_radii[atomic_numbers[atoms0[0]]]
    refn=covalent_radii[atomic_numbers[atoms0[-1]]]
    avg=(ref0+refn)*pvalue/float(circ1)
    #----------------------------------
    t, angle=float(0.0), 0.0
    si=atoms0.pop()
    helix=Atoms(symbols=[si], positions=[(float(radiomax), 0.0, 0.0)])
    last=[si]
    while atoms0:
        si=atoms0.pop()
        a1=covalent_radii[atomic_numbers[last[-1]]]
        a2=covalent_radii[atomic_numbers[si]]    
        S=(a1+a2)*pvalue
        theta=S/radiomax
        angle=angle+theta
        xx=radiomax*np.cos(angle)
        yy=radiomax*np.sin(angle)
        t=t+float(avg)
        inside=100
        while inside > 50:
            ai=Atom(symbol=si, position=(float(xx), float(yy), t))
            dmin=distance_atom2atoms(ai, helix)
            if (dmin >= ratio_radii_min) or (dmin >= ratio_radii_max): inside=0
            else: t = t - float(0.01)
        helix.append(ai)
        last.append(si)
    helix.info['e'] = 0.0
    helix.info['i'] = 'make_molecule_helix'
    return helix
#------------------------------------------------------------------------------------------
def make_molecule_eye(inatoms):
    atom_step=min_step(inatoms)
    atoms=list(inatoms)
    random.shuffle(atoms)
    si=atoms.pop()
    meyes=make_molecule_circle(atoms)
    zzz, inside=float(0.0), 100
    while inside > 50:
        tryatom=Atom(symbol=si, position=(float(0.0), float(0.0), zzz))        
        dmin=distance_atom2atoms(tryatom, meyes)
        if (dmin >= ratio_radii_min):
            meyes.append(tryatom)
            inside=0
        else: zzz = zzz + atom_step
    meyes.info['e'] = 0.0
    meyes.info['i'] = 'make_molecule_eye'
    return meyes
#------------------------------------------------------------------------------------------
def single_random_molecule_without_h(inatoms):
    if len(inatoms)==1:
        atoms=list(inatoms)
        m1=Atoms(symbols=[atoms[0]], positions=[(float(0.0), float(0.0), float(0.0))])
    else:
        which=random.randrange(0, 18)
        if which in [0, 1, 2]:    m1=make_molecule_3Da(inatoms)
        if which in [3, 4, 5]:    m1=make_molecule_3Db(inatoms)
        if which in [6, 7, 8, 9]: m1=make_molecule_2Da(inatoms)
        if which in [10, 11, 12]: m1=make_molecule_sphere(inatoms)
        if which in [13, 14]:     m1=make_molecule_helix(inatoms)
        if which == 15:           m1=make_molecule_wire(inatoms)
        if which == 16:           m1=make_molecule_circle(inatoms)
        if which == 17:           m1=make_molecule_two_rings(inatoms)
        if which == 18:           m1=make_molecule_eye(inatoms)
    return m1
#------------------------------------------------------------------------------------------
def add_atomrand(atomsin, newatomlist, randomatomplacement = 0, rec=0):            
    numlockedatoms, count, index = 0, 0, 0
    vecr=rand_unit_vector()
    last_success, mismatch = 0, 0
    start = random.randrange(0, len(atomsin))
    while(numlockedatoms < len(newatomlist) and count< 100*len(newatomlist)):        
        count+=1
        if(randomatomplacement == 1):
            index = random.randrange(0, len(atomsin))
        else:
            index = (numlockedatoms+start+int((count-last_success)/10)+mismatch)%len(atomsin)
        mainvector=atomsin[index].position
        vec=rand_unit_vector()
        norma=(covalent_radii[atomsin[index].number]+covalent_radii[atomic_numbers[newatomlist[numlockedatoms]]])*pvalue
        vecr = mainvector + norma*vec
        tryatom = Atom(symbol=newatomlist[numlockedatoms], position=vecr)        
        dmin=distance_atom2atoms(tryatom, atomsin)
        if (dmin >= ratio_radii_min):            
            if(numlockedatoms == 0):
                newatoms=Atoms(symbols=[tryatom.symbol], positions=[tryatom.position])
                numlockedatoms+=1    
            else:
                dmin=distance_atom2atoms(tryatom, newatoms)
                if (dmin > ratio_radii_max):                
                    newatoms.append(tryatom)                    
                    mismatch+=int((count-last_success)/10)
                    last_success=count    
                    numlockedatoms+=1        
    if(numlockedatoms>0):
        atomsout = merge_atoms([atomsin, newatoms])
    else:
        atomsout = atomsin.copy()

    ### If the process is not completed after 100 * (number of atoms to add) attempts,
    #   the function "add_atomrand" is called recursively 10 times using the final Atoms object.
    #   If the process still fails, an intentional failure is triggered,
    #   and the process is terminated.

    if(numlockedatoms<len(newatomlist)):
        print("FALL ... tries: %s" %(count))
        if(rec<10):
            senderlist=[]
            for ii in range(numlockedatoms, len(newatomlist)):
                senderlist.append(newatomlist[ii])
            atomsout = add_atomrand(atomsout, senderlist, randomatomplacement, rec+1)
        else:
            mainvector=atomsin[0].position
            vec=rand_unit_vector()        
            norma=(covalent_radii[atomsin[index].number]+covalent_radii[atomic_numbers[newatomlist[numlockedatoms]]])*(ratio_radii_min-0.1)
            vecr = mainvector + norma*vec            
            tryatom = Atom(symbol=newatomlist[numlockedatoms], position=vecr)        
            newatoms.append(tryatom)        
            print("QUIT add_atomrand()")    
    return atomsout
#------------------------------------------------------------------------------------------
def single_random_molecule(inatoms):
    ticklish, skeleton =[], []
    for iatom in inatoms:
        if iatom in reactivelist: ticklish.append(iatom)
        else:                     skeleton.append(iatom)
    moleculeout=single_random_molecule_without_h(skeleton)
    if ticklish != []: moleculeout=add_atomrand(moleculeout, ticklish)
    return moleculeout
#------------------------------------------------------------------------------------------
def make_molecules_random(inatoms, cuantas, index=0):    
    m0, count=[], 0
    for ii in range(cuantas):
        m1=single_random_molecule(inatoms)
        rmin=distance_interatom(m1)
        if rmin >= ratio_radii_min:
            ans=connected_graph(m1, ratio_radii_max)
            if ans:
                count=count+1
                m0.extend([m1])
                if(len(m0) == cuantas): break             
    #print("Generando aleatoriamente %s de %s mol (proc: %s)" %(count, cuantas, index))   
    return m0
#------------------------------------------------------------------------------------------



import os
import numpy as np
from ase.data import covalent_radii
from ase.data.colors import cpk_colors, jmol_colors
# ------------------------------------------------------------------------------------------
# Visualization parameters
bwd = 0.10                     # Bond thickness
dlv = 0.15                     # Cell edge thickness
sphere_factor = 0.60           # Atom size scale factor
projection = 'orthographic'    # or 'perspective'
reflection = 'reflection 0.0'
reflection_model = 'phong 1.0' # or 'specular'
radio_factor = 1.30            # Bond length scaling factor
tmit = 0.00                    # Transparency (transmit parameter)
color_scheme = 'jmol'          # 'cpk' or 'jmol'
# ------------------------------------------------------------------------------------------
def get_color(Z):
    """Return atomic color according to the selected color scheme."""
    return cpk_colors[Z] if color_scheme == 'cpk' else jmol_colors[Z]
# ------------------------------------------------------------------------------------------
aspect_ratios_4by3 = {
    'a': (800, 600),
    'b': (1024, 768),
    'c': (1280, 960),
    'd': (1600, 1200),
    'e': (2048, 1536),
    'f': (3200, 2400),
    'g': (6000, 4500),
}
# ------------------------------------------------------------------------------------------
def draw_arrows(atoms, forces, opnew):
    try:
        arrow_scale=10.0
        arrow_thickness=0.05
        arrow_color=(0, 0, 0)
        arrow_rgb = f"<{arrow_color[0]:.3f},{arrow_color[1]:.3f},{arrow_color[2]:.3f}>"
        for pos, F in zip(atoms.positions, forces):
            normF = np.linalg.norm(F)
            if normF < 1e-6:
                continue
            uF = F / normF
            shaft_length = normF * arrow_scale * 0.85
            cone_length = normF * arrow_scale * 0.15
            start = pos
            end_shaft = pos + uF * shaft_length
            end_cone = end_shaft + uF * cone_length

            # Draw arrow shaft (cylinder)
            print(
                f'cylinder {{<{start[0]:.6f},{start[1]:.6f},{start[2]:.6f}> '
                f'<{end_shaft[0]:.6f},{end_shaft[1]:.6f},{end_shaft[2]:.6f}>, {arrow_thickness:.3f} '
                f'pigment{{color rgb {arrow_rgb}}} finish {{{reflection_model}}}}}',
                file=opnew
            )
            # Draw arrow cone
            print(
                f'cone {{<{end_shaft[0]:.6f},{end_shaft[1]:.6f},{end_shaft[2]:.6f}>, {arrow_thickness*2:.3f} '
                f'<{end_cone[0]:.6f},{end_cone[1]:.6f},{end_cone[2]:.6f}>, 0 '
                f'pigment{{color rgb {arrow_rgb}}} finish {{{reflection_model}}}}}',
                file=opnew
            )
    except Exception as e:
        print(f"Could not draw force arrows: {e}")
# ------------------------------------------------------------------------------------------
def draw_cell(atoms, opnew):
    """Draw the unit cell edges as black cylinders for periodic systems."""
    matrix = atoms.cell
    a1, a2, a3 = matrix
    a0 = -(a1 + a2 + a3) / 2
    edges = [
        (a0, a0 + a1), (a0, a0 + a2), (a0, a0 + a3),
        (a0 + a1, a0 + a1 + a2), (a0 + a1, a0 + a1 + a3),
        (a0 + a2, a0 + a1 + a2), (a0 + a2, a0 + a2 + a3),
        (a0 + a3, a0 + a1 + a3), (a0 + a3, a0 + a2 + a3),
        (a0 + a1 + a2, a0 + a1 + a2 + a3),
        (a0 + a1 + a3, a0 + a1 + a2 + a3),
        (a0 + a2 + a3, a0 + a1 + a2 + a3)
    ]
    for v1, v2 in edges:
        x1, y1, z1 = v1
        x2, y2, z2 = v2
        print(
            f'cylinder {{<{x1:9.6f},{y1:9.6f},{z1:9.6f}> '
            f'<{x2:9.6f},{y2:9.6f},{z2:9.6f}>, {dlv:.4f} '
            f'pigment{{color rgb <0,0,0>}} finish {{{reflection_model} {reflection}}}}}',
            file=opnew
        )
    d = max(np.linalg.norm(v) for v in [a1, a2, a3])
    return d, a0
# ------------------------------------------------------------------------------------------
def write_image(atoms, basename, f=1.2, quality='b', draw_forces=False):
    """Generate a PNG image using POV-Ray, optionally drawing force arrows."""

    width, height = aspect_ratios_4by3.get(quality, (1024, 768))
    name_pov = basename + '.pov'
    name_png = basename + '.png'

    forces = None
    if 'forces' in atoms.arrays:
        forces = atoms.arrays['forces']
    elif atoms.calc is not None:
        try:
            forces = atoms.get_forces()
        except Exception:
            forces = None

    # Open POV-Ray scene file
    with open(name_pov, 'w') as opnew:
        print('global_settings { assumed_gamma 1.0 }', file=opnew)
        print('background { color rgb<2.0, 2.0, 2.0> }', file=opnew)
        print('light_source { <10, -8, -8> color rgb <1,1,1> }', file=opnew)
        print('light_source { <-8, 8, 8> color rgb <1,1,1> }', file=opnew)

        # Check if any periodic boundary condition is True
        if atoms.pbc.any():
            d,a0=draw_cell(atoms, opnew)
            atoms.translate(a0)
        else:
            # Original camera distance calculation for non-periodic systems
            ctd = atoms.get_center_of_mass()
            r = [np.linalg.norm(atom.position - ctd) + covalent_radii[atom.number] for atom in atoms]
            d = 2.0 * max(r)

        # Camera
        factor = f * d
        print(f'camera {{{projection} location <0,0,{factor:.2f}> look_at <0,0,0> rotate y*0.0}}\n', file=opnew)

        # Bonds
        nn = len(atoms)
        for i in range(nn):
            ri = covalent_radii[atoms[i].number]
            pos_i = atoms[i].position
            for j in range(i + 1, nn):
                rj = covalent_radii[atoms[j].number]
                pos_j = atoms[j].position
                uij = pos_j - pos_i
                rr = np.linalg.norm(uij)
                if rr == 0:
                    continue
                if rr / (ri + rj) < radio_factor:
                    uijn = uij / rr
                    col_i = get_color(atoms[i].number)
                    col_j = get_color(atoms[j].number)
                    start_i = pos_i + 0.5 * ri * uijn
                    start_j = pos_j - 0.5 * rj * uijn
                    mid = (start_i + start_j) / 2
                    for (xa, ya, za, xb, yb, zb, col) in [
                        (start_i[0], start_i[1], start_i[2], mid[0], mid[1], mid[2], col_i),
                        (mid[0], mid[1], mid[2], start_j[0], start_j[1], start_j[2], col_j)
                    ]:
                        kolor = f"<{col[0]:.3f},{col[1]:.3f},{col[2]:.3f}>"
                        print(
                            f'cylinder {{<{xa:9.6f},{ya:9.6f},{za:9.6f}> '
                            f'<{xb:9.6f},{yb:9.6f},{zb:9.6f}>, {bwd:.2f} '
                            f'pigment {{color rgb {kolor} transmit {tmit:.2f}}} '
                            f'finish {{{reflection_model} {reflection}}}}}',
                            file=opnew
                        )
        # Atoms
        for atom in atoms:
            ri = covalent_radii[atom.number]
            cpk = get_color(atom.number)
            xx, yy, zz = atom.position
            kolor = f"<{cpk[0]:.3f},{cpk[1]:.3f},{cpk[2]:.3f}>"
            print(
                f'sphere {{<{xx:9.6f},{yy:9.6f},{zz:9.6f}>, {sphere_factor*ri:6.4f} '
                f'pigment {{color rgb {kolor} transmit {tmit:.2f}}} '
                f'finish {{{reflection_model} {reflection}}}}}',
                file=opnew
            )

        if draw_forces and forces is not None:
            draw_arrows(atoms, forces, opnew)

    os.system(f"povray +A Display=Off Output_File_Type=N Width={width} Height={height} {name_pov} > /dev/null 2>&1")
    os.system(f"rm -f {name_pov}")
# ------------------------------------------------------------------------------------------
#import os
#import numpy as np
#from ase.io import read
#from ase.calculators.singlepoint import SinglePointCalculator
#from aegon.libatoms2png import write_image
# ------------------------------------------------------------------------------------------
#traj='rand01.xyz'
#frames = read(traj, index=':')
#for i, frame in enumerate(frames):
#    basename = f"frame{i:03d}"
#    F = frame.get_forces()
#    fmax = float(np.max(np.linalg.norm(F, axis=1)))
#    print(f"{basename} fmax={fmax:8.6f}")
#    frame.calc = SinglePointCalculator(frame, forces=F)
#    write_image(frame, basename, f=1.2, quality='b', draw_forces=True)
#os.system('magick -delay 10 -loop 0 frame*.png caso.gif')
#os.system("rm -f frame*.png")
# ------------------------------------------------------------------------------------------
#atoms = read('poscar001.vasp')
#write_image(atoms, 'crystal', f=1.2, quality='b', draw_forces=False)



import random
import numpy as np
from ase import Atom, Atoms
from aegon.libutils import rand_unit_vector, rotate_random, distance_atoms2atoms, rotate_vector_angle_deg, rename
#------------------------------------------------------------------------------------------
""" CUT IN THE Z AXIS """
def get_gen(atoms):
    ene=float(0.0)
    list_up, list_dn = [], []
    children_up=Atoms()
    children_dn=Atoms()
    for ii, iatom in enumerate(atoms):
        if iatom.position[2] >= 0.0:
            children_up.append(iatom)
            list_up.append(ii)
        else:
            children_dn.append(iatom)
            list_dn.append(ii)
    children_up.info['c']=list_up
    children_dn.info['c']=list_dn
    return children_up, children_dn
#------------------------------------------------------------------------------------------
def randvector(r):
    ru=float(random.uniform(0.0, r))
    vector = ru*rand_unit_vector()
    return vector
#------------------------------------------------------------------------------------------
def molecule_from_list(moleculein, n):
    if not isinstance(n, list): n = [n]
    moleculeout = Atoms()
    for i in n: moleculeout.append(moleculein[i])
    return moleculeout
#------------------------------------------------------------------------------------------
def up_part_from_n(father1, nff, deltadisp):
    fall=True
    contador=0
    while fall:
        father2=father1.copy()
        rv=randvector(deltadisp)
        father2.translate(-rv)
        rotate_random(father2)
        fup, fdn=get_gen(father2)
        fup_cs=fup.get_chemical_symbols()
        fdn_cs=fdn.get_chemical_symbols()
        if len(fup_cs)==nff:
            fall=False
            return fup, fdn
        elif len(fdn_cs)==nff:
            fall=False
            positions = fup.get_positions()
            positions[:, 2] *= -1
            fup.set_positions(positions)
            positions = fdn.get_positions()
            positions[:, 2] *= -1
            fdn.set_positions(positions)
            return fdn, fup
        else:
            contador=contador+1
            if contador==5000:
                return False
#------------------------------------------------------------------------------------------
def dn_part_from_zlist(mother1, list_chemical_symbols, deltadisp):
    fall=True
    contador=0
    while fall:
        mother2=mother1.copy()
        rv=randvector(deltadisp)
        mother2.translate(-rv)
        rotate_random(mother2)
        mup, mdn=get_gen(mother2)
        mdn_cs=mdn.get_chemical_symbols()
        mup_cs=mup.get_chemical_symbols()
        if (mdn_cs==list_chemical_symbols):
            fall=False
            return mup, mdn
        elif (mup_cs==list_chemical_symbols):
            fall=False
            positions = mup.get_positions()
            positions[:, 2] *= -1
            mup.set_positions(positions)
            positions = mdn.get_positions()
            positions[:, 2] *= -1
            mdn.set_positions(positions)
            return mdn, mup
        else:
            contador=contador+1
            if contador==5000:
                return False
#------------------------------------------------------------------------------------------
def mol_min_radius(atoms):
    moleculeout=atoms.copy()
    moleculeout.translate(-moleculeout.get_center_of_mass())
    positions = atoms.get_positions()
    distances = np.linalg.norm(positions, axis=1)
    distances.sort()
    rmin=distances[0]*float(0.6)
    return rmin
#------------------------------------------------------------------------------------------
def complement(tot, part):
    ans=tot.copy()
    for elemi in part:
      if elemi in ans: ans.remove(elemi)
    return(ans)
#------------------------------------------------------------------------------------------
def cut_deavenho(singlefather, singlemother, atomlist):
    father1=singlefather.copy()
    mother1=singlemother.copy()
    father1.translate(-father1.get_center_of_mass())
    mother1.translate(-mother1.get_center_of_mass())
    rminf=mol_min_radius(father1)
    rminm=mol_min_radius(mother1)
    n0=len(atomlist)//2
    for ii, nff in enumerate([n0, n0-1, n0+1]):
        ans1=up_part_from_n(father1, nff, rminf)
        if (ans1 is False):
            return False
        fup, fdn  = ans1
        acfzlist=fup.get_chemical_symbols()
        rsdzlist=complement(atomlist, acfzlist)
        ans2=dn_part_from_zlist(mother1, rsdzlist, rminm)
        if (ans2 is False):
            if (ii == 2):
                return False
            else:
                continue
        else:
            mup, mdn = ans2
            break
    moleculeout=[fup, mdn]
    return moleculeout
#------------------------------------------------------------------------------------------
def crossover_deavenho(singlefather, singlemother, atomlist):
    moleculelist=Atoms()
    father=singlefather.copy()
    mother=singlemother.copy()
    result=cut_deavenho(singlefather, singlemother, atomlist)
    if not result:
        result=cut_deavenho(singlemother, singlefather, atomlist)
    if not result:
        return False
    else:
        gena,genb=result
        ratio_radii_crit=0.90
        var=gena.copy()
        moleculelist=[var+genb]
        dmin=distance_atoms2atoms(var, genb)
        test = False if (dmin < ratio_radii_crit) else True
        angle=0
        zeta=[0.0,0.0,1.0]
        vz=np.array([float(0.0), float(0.0), float(0.05)])
        while test == False:
            angle=angle+10
            var=rotate_vector_angle_deg(gena, zeta, angle)
            moleculelist.append(var+genb)
            dmin=distance_atoms2atoms(var, genb)
            if (dmin >= ratio_radii_crit):
                #writexyzs(moleculelist,'fenomeno.xyz')
                test=True
            if (angle >=360):
                angle=0
                gena.translate(-vz)
        moleculeout=moleculelist[-1]
        return moleculeout
#------------------------------------------------------------------------------------------
def make_children(papy, mamy, atomlist):
    moleculeout=[]
    total_molecules=len(papy)
    for ii in range(total_molecules):
        ichildren=crossover_deavenho(papy[ii], mamy[ii], atomlist)
        if ichildren is not False:
            ichildren.e=float(0.0)
            moleculeout.extend([ichildren])
        else:
            print("make_children(%s, %s) FALL" %(papy[ii].i,mamy[ii].i))
    moleculeout=rename(moleculeout, 'pre_child_', 4)
    return moleculeout
#------------------------------------------------------------------------------------------




import queue
import time
import subprocess
from multiprocessing import Process, Queue, current_process
# ------------------------------------------------------------------------------------------
import warnings
import logging
# Configuración de advertencias y logging
warnings.filterwarnings("ignore")
logging.captureWarnings(True)
# ------------------------------------------------------------------------------------------
def execute_bash(tasks_to_accomplish, tasks_that_are_done):
    """Ejecuta archivos de ordenes bash en paralelo."""
    while True:
        try:
            task = tasks_to_accomplish.get_nowait()
        except queue.Empty:
            break
        else:
            try:
                subprocess.run(f"bash {task}", shell=True, check=True)
                tasks_that_are_done.put(f"{task} calculated by {current_process().name}")
            except subprocess.CalledProcessError as e:
                print(f"Error executing bash for {task}: {e}")
            time.sleep(0.01)  # Pequeña pausa para evitar sobrecargar el sistema

# ------------------------------------------------------------------------------------------
def parallel_bash_execution(list_of_sh_files, NparCalcs=4):
    """Ejecuta cálculos en paralelo usando multiprocessing."""
    tasks_to_accomplish = Queue()
    tasks_that_are_done = Queue()
    processes = []

    for sh_file in list_of_sh_files:
        tasks_to_accomplish.put(sh_file)

    for _ in range(NparCalcs):
        p = Process(target=execute_bash, args=(tasks_to_accomplish, tasks_that_are_done))
        processes.append(p)
        p.start()

    for p in processes:
        p.join()

    while not tasks_that_are_done.empty():
        print(tasks_that_are_done.get())

# ------------------------------------------------------------------------------------------
#import glob
#cases_list=['stage2'+str(i+1).zfill(5) for i in range(45)]
#output_files = sorted(glob.glob("stage2[0-9][0-9][0-9][0-9][0-9].out"))
#shfiles_to_send = [ibname+'.sh' for ibname in cases_list if ibname+'.out' not in output_files]
#parallel_bash_execution(shfiles_to_send, NparCalcs=2)



import glob
import os.path
import numpy as np
from ase import Atom, Atoms
#------------------------------------------------------------------------------------------
eVtokcalpermol = 23.060548012069496
#------------------------------------------------------------------------------------------
def get_normaltermination_mopac(filename):
    if not os.path.isfile(filename):
        return -1
    abnormal = 0
    with open(filename, 'r') as f:
        for line in f:
            if "Error and normal termination messages reported" in line:
                abnormal += 1
    return 1 if abnormal == 0 else 0
#------------------------------------------------------------------------------------------
def get_energy_mopac(filename):
    ene_eV = None
    with open(filename, 'r') as f:
        for line in f:
            if "TOTAL ENERGY" in line:
                try:
                    ene_eV = float(line.split()[3])
                except (IndexError, ValueError):
                    continue
    if ene_eV is None:
        return None
    return ene_eV * eVtokcalpermol
#------------------------------------------------------------------------------------------
def get_cell_mopac(filename):
    """Return 3x3 cell matrix if present, otherwise None"""
    with open(filename, 'r') as f:
        for line in f:
            if '      Tv         ' in line:
                a = []
                for _ in range(3):
                    ls = next(f).split()
                    if len(ls) < 6:
                        a.append([float(ls[2]), float(ls[3]), float(ls[4])])
                    else:
                        a.append([float(ls[2]), float(ls[4]), float(ls[6])])
                return np.array(a)
    return None
#------------------------------------------------------------------------------------------
def get_geometry_mopac(pathfilename):
    nt = get_normaltermination_mopac(pathfilename)
    if nt == -1:
        return False
    energy = get_energy_mopac(pathfilename)
    if energy is None:
        return False
    namein = os.path.basename(pathfilename).split('.')[0]
    cell = get_cell_mopac(pathfilename)
    atoms = Atoms()
    atoms.info['c'] = nt
    atoms.info['e'] = energy
    atoms.info['i'] = namein
    if cell is not None:
        atoms.set_cell(cell)
        atoms.pbc = True
    with open(pathfilename, 'r') as f:
        for line in f:
            if '                             CARTESIAN COORDINATES' in line:
                for ii in range(2): line=f.readline()
                ls = line.split()
                while ( ls != [] ):
                    symbol=str(ls[1])
                    x, y, z=float(ls[2]), float(ls[3]), float(ls[4])
                    atoms.append(Atom(symbol=symbol, position=(x, y, z)))
                    line=f.readline()
                    ls = line.split()
    return atoms
#------------------------------------------------------------------------------------------
#------------------------------------------------------------------------------------------
def xyzs_from_outputs_mopac():
    molecules = []
    for outfile in sorted(glob.glob("*.out")):
        mol = get_geometry_mopac(outfile)
        if mol is not False:
            molecules.append(mol)
        else:
            print(f"Skipped {outfile}")
    return molecules
#------------------------------------------------------------------------------------------
def get_freqneg_mopac(filename):
    freq_negative = 0
    with open(filename, 'r') as f:
        for line in f:
            if "FREQ." in line:
                freq = float(line.split()[1])
                if freq < 0.0:
                    freq_negative += 1
    return freq_negative
#------------------------------------------------------------------------------------------
def cutter_negfreq_mopac(moleculein, path):
    moleculeout = []
    for mol in moleculein:
        basename = mol.info['i']
        freqneg = get_freqneg_mopac(os.path.join(path, basename + '.out'))
        if freqneg == 0:
            moleculeout.append(mol)
        else:
            print(f"{basename} ... DISCARDED (negative frequency)")
    return moleculeout
#------------------------------------------------------------------------------------------
def example()
    from aegon.libutils import writexyzs
    filename='pit.out'
    nt=get_normaltermination_mopac(filename)
    ene=get_energy_mopac(filename)
    print("Etot = %.8f kcal/mol" %(ene))
    atom=get_geometry_mopac(filename)
    writexyzs(atom, 'final.xyz')



import random
import numpy as np
from aegon.libutils import sort_by_energy
#-------------------------------------------------------------------------------
def get_fitness(moleculelist):
    if len(moleculelist)==1:
       fitness=[float(1.0)]
    else:
       fitness=[]
       listmolecule=sort_by_energy(moleculelist,1)
       Emin=listmolecule[0].info['e']
       Emax=listmolecule[-1].info['e']
       for imol in listmolecule:
           Ei=imol.info['e']
           EFEi=(Ei-Emin)/(Emax-Emin)
           fi=0.5*(1.0-np.tanh(((2.0*EFEi)-1.0)))
           fitness.append(float(fi))
    return fitness
#-------------------------------------------------------------------------------
def get_roulette_wheel_selection(moleculelist, nmating):
    listmolecule=sort_by_energy(moleculelist,1)
    fitness=get_fitness(listmolecule)
    sum_of_fitness=sum(fitness)
    previous_probability=0.0
    n=len(listmolecule)
    pp=[]
    for ix in range(n):
        previous_probability= previous_probability+ (fitness[ix]/sum_of_fitness)
        if ix==n-1: pp.append(1.0)
        else: pp.append(previous_probability)
    roulette = []
    for i in range(nmating):
        random_number = random.random()
        for ii, p in enumerate(pp):
            if random_number <= p:
                roulette.append(listmolecule[ii])
                break
    return roulette
#-------------------------------------------------------------------------------



import time
import numpy as np
from joblib import Parallel, delayed
from scipy.optimize import minimize
from numba import jit
#------------------------------------------------------------------------------------------
cutoff=18.0
method='L-BFGS-B'
gtol=1e-6
#------------------------------------------------------------------------------------------
#PHILOSOPHICAL MAGAZINE LETTERS, 1990, VOL. 61, No. 3, 139-146
SUTTON_CHEN_PARAMS = {
    'Ni': {'n': 9,  'm': 6, 'epsilon': 1.5707E-02, 'af': 3.52, 'c': 39.432},
    'Cu': {'n': 9,  'm': 6, 'epsilon': 1.2382E-02, 'af': 3.61, 'c': 39.432},
    'Rh': {'n': 12, 'm': 6, 'epsilon': 4.9371E-03, 'af': 3.80, 'c': 144.41},
    'Pd': {'n': 12, 'm': 7, 'epsilon': 4.1790E-03, 'af': 3.89, 'c': 108.27},
    'Ag': {'n': 12, 'm': 6, 'epsilon': 2.5415E-03, 'af': 4.09, 'c': 144.41},
    'Ir': {'n': 14, 'm': 6, 'epsilon': 2.4489E-03, 'af': 3.84, 'c': 334.94},
    'Pt': {'n': 10, 'm': 8, 'epsilon': 1.9833E-02, 'af': 3.92, 'c': 34.408},
    'Au': {'n': 10, 'm': 8, 'epsilon': 1.2793E-02, 'af': 4.08, 'c': 34.408},
    'Pb': {'n': 10, 'm': 7, 'epsilon': 5.5765E-03, 'af': 4.95, 'c': 45.778},
    'Al': {'n': 7,  'm': 6, 'epsilon': 3.3147E-02, 'af': 4.05, 'c': 16.399}
}
#------------------------------------------------------------------------------------------
@jit(nopython=True, fastmath=True)
def opt_sc_metal(positions, n, m, epsilon, af, c):
    """Calculate Sutton-Chen energy (optimized with Numba)"""
    N = len(positions)
    s1 = np.zeros(N)
    s2 = np.zeros(N)
    energy = 0.0
    for i in range(N - 1):
        for j in range(i + 1, N):
            rij = positions[i] - positions[j]
            r = np.linalg.norm(rij)
            if r < cutoff:
                invr = af / r
                var0n = invr ** n
                var0m = invr ** m
                s1[i] += var0n
                s1[j] += var0n
                s2[i] += var0m
                s2[j] += var0m
    energy = epsilon * (0.5 * np.sum(s1) - c * np.sum(np.sqrt(s2)))
    return energy
#------------------------------------------------------------------------------------------
#https://wanglab.hosted.uark.edu/DLPOLY2/node56.html
@jit(nopython=True, fastmath=True)
def opt_sc_forces(positions, n, m, epsilon, af, c):
    N = len(positions)
    forces = np.zeros_like(positions)
    s2 = np.zeros(N)
    for i in range(N - 1):
        for j in range(i + 1, N):
            rij = positions[i] - positions[j]
            r = np.linalg.norm(rij)
            if r < cutoff:
                var0 = af / r
                var0_m = var0 ** m
                s2[i] += var0_m
                s2[j] += var0_m
    sqrt_rho = np.sqrt(s2)
    for i in range(N - 1):
        for j in range(i + 1, N):
            rij = positions[i] - positions[j]
            r = np.linalg.norm(rij)
            if r < cutoff:
                var0 = af / r
                var0_n = var0 ** n
                var0_m = var0 ** m
                f_rep = epsilon * n * var0_n / r
                f_dens_i = epsilon * c * m * var0_m / (2.0 * sqrt_rho[i] * r)
                f_dens_j = epsilon * c * m * var0_m / (2.0 * sqrt_rho[j] * r)
                f_scalar = f_rep - f_dens_i - f_dens_j
                fij = f_scalar * (rij / r)
                forces[i] += fij
                forces[j] -= fij
    return forces
#------------------------------------------------------------------------------------------
def sutton_chen_energy(atoms, metal_type):
    params = SUTTON_CHEN_PARAMS[metal_type]
    positions = atoms.get_positions()
    return opt_sc_metal(
        positions, 
        params['n'], 
        params['m'], 
        params['epsilon'], 
        params['af'], 
        params['c']
    )
#------------------------------------------------------------------------------------------
def sutton_chen_forces(atoms, metal_type):
    """Wrapper para fuerzas Sutton-Chen con objetos ASE"""
    params = SUTTON_CHEN_PARAMS[metal_type]
    positions = atoms.get_positions()
    return opt_sc_forces(
        positions, 
        params['n'], 
        params['m'], 
        params['epsilon'], 
        params['af'], 
        params['c']
    )
#------------------------------------------------------------------------------------------
def opt_sc(atoms, metal_type='Cu'):
    params = SUTTON_CHEN_PARAMS[metal_type]
    af = params['af']
    n = params['n']
    m = params['m']
    epsilon = params['epsilon']
    c = params['c']
    cluster = atoms.copy()
    x0 = cluster.get_positions()
    def objective(x):
        positions = x.reshape(-1, 3)
        return opt_sc_metal(positions, n, m, epsilon, af, c)
    def gradient(x):
        positions = x.reshape(-1, 3)
        forces = opt_sc_forces(positions, n, m, epsilon, af, c)
        return -forces.reshape(-1)  # Gradiente = -Fuerzas
    result = minimize(
        fun=objective,
        x0=x0.reshape(-1),
        jac=gradient,
        method=method,
        options={'disp': False, 'gtol': gtol}
    )
    opt_positions = result.x.reshape(-1, 3)
    energy_opt = result.fun
    cluster_opt = atoms.copy()
    cluster_opt.set_positions(opt_positions)
    cluster_opt.info['e'] = energy_opt
    cluster_opt.info['c'] = 1 if result.success else 0
    return cluster_opt
#------------------------------------------------------------------------------------------
def opt_SC_parallel(mol_list, metal_type='Cu', n_jobs=-1):
    start_time = time.time()
    n1=len(mol_list)
    if not isinstance(mol_list, list): mol_list = [mol_list]
    results = Parallel(n_jobs=n_jobs)(delayed(opt_sc)(mol, metal_type) for mol in mol_list)
    n2=len(results)
    end_time = time.time()
    print("Local OPT parallel at %.2f s [%d -> %d]" % (end_time-start_time, n1, n2))
    return results
#------------------------------------------------------------------------------------------
#from aegon.libutils import readxyzs
#listmol=readxyzs('../SC052.xyz')
#start_time = time.time()
#for imol in listmol: e=sutton_chen_energy(imol, metal_type='Pd')
##print(e)
#end_time = time.time()
#print("Referencia at %.3f s" % (end_time-start_time))
#start_time = time.time()
#for imol in listmol: e=sutton_chen_forces(imol, metal_type='Pd')
##print(e)
#end_time = time.time()
#print("Propuesta at %.3f s" % (end_time-start_time))



import os.path
import numpy as np
from ase.io import read
from ase import Atom, Atoms
#------------------------------------------------------------------------------------------
def get_termination_vasp(pathfilename):
    if os.path.isfile(pathfilename):
        file=open(pathfilename,'r')
        normal=0
        for line in file:
            if "General timing and accounting informations for this job" in line: normal=normal+1
        file.close()
        return normal
    else:
        return False
#------------------------------------------------------------------------------------------
#Provides the minimum energy structure of the optimization process, regardless of whether it converged correctly
def get_geometry_vasp_old(pathfilename):
    nt=get_termination_vasp(pathfilename)
    if nt==False: return False
    filename=os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    trajectory = read(pathfilename, index=':', format='vasp-out')
    enemin = float('inf')
    for i, atoms in enumerate(trajectory):
        energy=atoms.get_potential_energy()
        if energy < enemin:
            enemin=energy
            poscar=atoms.copy()
            poscar.info['c']=nt
            poscar.info['e']=energy
            poscar.info['i']=namein
    return poscar
#------------------------------------------------------------------------------------------
def get_geometry_vasp(pathfilename):
    nt=get_termination_vasp(pathfilename)
    if nt==False: return False
    filename=os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    trajectory = get_traj_vasp(pathfilename,False)
    enemin = float('inf')
    for i, atoms in enumerate(trajectory):
        energy=atoms.info['e']
        if energy < enemin:
            enemin=energy
            poscar=atoms.copy()
            poscar.info['c']=nt
            poscar.info['e']=energy
            poscar.info['i']=namein
    return poscar
#------------------------------------------------------------------------------------------
def get_traj_vasp_old(pathfilename, force=False):
    nt=get_termination_vasp(pathfilename)
    if nt==False: return False
    filename=os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    trajectory = read(pathfilename, index=':', format='vasp-out')
    atoms_list_out=[]
    for i, atoms in enumerate(trajectory):
        if force: atoms.arrays['forces']=atoms.get_forces()
        energy=atoms.get_potential_energy()
        atoms.info['c']=nt
        atoms.info['e']=energy
        atoms.info['i']=namein+str(i+1).zfill(3)
        atoms_list_out.extend([atoms])
    return atoms_list_out
#------------------------------------------------------------------------------------------
def get_traj_vasp(filename, force=False):
    namein=filename.split('.')[0]
    start, end, ene, elements,counts,matrices  = [], [], [], [], [], []
    openold = open(filename,"r")
    rline = openold.readlines()
    for i in range(len(rline)):
        if "direct lattice vectors" in rline[i]:
            matrix = np.zeros((3, 3))
            matrix[0,0],matrix[0,1],matrix[0,2]= float(rline[i+1].split()[0]),float(rline[i+1].split()[1]),float(rline[i+1].split()[2])
            matrix[1,0],matrix[1,1],matrix[1,2]= float(rline[i+2].split()[0]),float(rline[i+2].split()[1]),float(rline[i+2].split()[2])
            matrix[2,0],matrix[2,1],matrix[2,2] = float(rline[i+3].split()[0]),float(rline[i+3].split()[1]),float(rline[i+3].split()[2])
            matrices.append(matrix)
        elif "POSITION" in rline[i]:
            start.append(i+2)
            for j in range(i + 2, len(rline)):
                if set(rline[j].strip()) == {"-"}:
                    end.append(j - 1)
                    break
        elif "FREE ENERGIE OF THE ION-ELECTRON SYSTEM (eV)" in rline[i]:
            ene.append(rline[i+2].split()[4])  # ELECTROVOLTS -EV
        elif "VRHFIN =" in rline[i]:
            elements.append(str(rline[i].split('=')[1].split(':')[0].strip()))
        elif "ions per type =" in rline[i]:
            counts = list(map(int, rline[i].split('=')[1].split()))
    elements_per_atoms = []
    for elem, count in zip(elements, counts):
        elements_per_atoms.extend([elem] * count)
    moleculeout=[]
    for i,iStart in enumerate(start):
        singlemol = Atoms()
        singlemol.info['e'] = float(ene[i])
        singlemol.info['i'] = namein+'_'+str(i+1).zfill(3)
        singlemol.cell = matrices[i]
        singlemol.pbc = True
        atom_index = 0
        for line in rline[start[i] : end[i]+1]:
            words = line.split()
            ss =  elements_per_atoms[atom_index]
            xc,yc,zc = float(words[0]), float(words[1]), float(words[2])
            ai=Atom(symbol=ss,position=(xc, yc, zc))
            singlemol.append(ai)
            atom_index += 1
        if force:
            forces_list_by_group = []
            for line in rline[start[i] : end[i]+1]:
                words = line.split()
                fx,fy,fz = float(words[3]), float(words[4]), float(words[5]) #eV/Angst
                forces_list_by_group.append([fx,fy,fz])
            singlemol.arrays['forces'] = np.array(forces_list_by_group)
        moleculeout.extend([singlemol])
    openold.close()
    return(moleculeout)
#------------------------------------------------------------------------------------------



from ase     import Atoms
from ase.io  import write as ase_write
from ase.calculators.singlepoint import SinglePointCalculator
from aegon.libcodegaussian import get_geometry_gaussian, get_traj_gaussian
from aegon.libcodeorca     import get_geometry_orca,     get_traj_orca
from aegon.libcodevasp     import get_geometry_vasp,     get_traj_vasp 
from aegon.libcodegulp     import get_geometry_gulp
from aegon.libcfg          import writecfgs
from aegon.libutils        import writexyzs
from aegon.libposcar       import writeposcars
#------------------------------------------------------------------------------------------
# CREATION OF CLASSES
class read_out:
    
    def __init__(self):
       self.fileout_traj = {
           "gaussian": get_traj_gaussian,
           "orca":     get_traj_orca,
           "vasp":     get_traj_vasp
       } 
       self.fileout_geometry = {
           "gaussian": get_geometry_gaussian,
           "orca":     get_geometry_orca,
           "vasp":     get_geometry_vasp,
           "gulp":     get_geometry_gulp
       } 
    def traj(self, document, file_name, force=False):
        if document not in self.fileout_traj:
            raise ValueError(f"This '{document}' do not exist.")
        if not isinstance(file_name, str):
            raise TypeError("The second argument must be a (str).")
        if not isinstance(force, bool):
            raise TypeError("The third argument must be a boolean (True/False).")
        return self.fileout_traj[document](file_name, force)
    
    def geo(self,document,name):
        if document not in self.fileout_geometry:
            raise ValueError(f"This '{document}' do not exist.")
        if not isinstance(name, str):
            raise TypeError("The second argument must be a (str).")   
        return self.fileout_geometry[document](name)
#------------------------------------------------------------------------------------------
class write:
    @classmethod
    def cfg(cls, data, output_file, force=False):
        try:
            if isinstance(data, bool) and not data:
                raise ValueError("Data is False. Cannot execute cfg.")
            else:
               if not isinstance(data, list):
                   data = [data]
            if not isinstance(output_file, str):
                raise ValueError("The first argument must be a (str).")
            if not isinstance(force, bool):
                raise ValueError("The second argument must be a boolean (True/False).")
            writecfgs(data, output_file, force)  
            print("Writing %s" %(output_file))
        except Exception as e:
            print(f"Error in cfg: {e}")
    @classmethod
    def xyz(cls, data, output_file):
        try:
            if isinstance(data, bool) and not data:
                raise ValueError("Data is False. Cannot execute cfg.")
            else:
                if not isinstance(data, list):
                    data = [data]
            if not isinstance(output_file, str):
                raise ValueError("The first argument must be a (str).")
            writexyzs(data, output_file)  
            print("Writing %s" %(output_file))
        except Exception as e:
            print(f"Error in xyz: {e}")
    @classmethod       
    def poscar(cls,data, output_file, opt='D'):
        try:
            if isinstance(data, bool) and not data:
                raise ValueError("Data is False. Cannot execute cfg.")
            else:
                if not isinstance(data, list):
                    data = [data]
            if not isinstance(output_file, str):
                raise ValueError("The first argument must be a (str).")
            if not isinstance(opt, str):
                raise ValueError("The second argument must be a (str). \tOptions:D,C or other.")
            writeposcars(data, output_file, opt)
            print("Writing %s" %(output_file))
        except Exception as e:
            print(f"Error in vasp: {e}")
    @classmethod
    def extxyz(cls, data, output_file, force=False):
        try:
            if isinstance(data, bool) and not data:
                raise ValueError("Data is False. Cannot execute cfg.")
            else:
                if not isinstance(data, list):
                    data = [data]
            if not isinstance(output_file, str):
                raise ValueError("The first argument must be a (str).")

            poscarout=[]
            for ipos in data:
                if any(ipos.pbc):
                    atoms = Atoms(ipos.symbols, positions=ipos.positions, cell=ipos.cell, pbc=True)
                else:
                    atoms = Atoms(ipos.symbols, positions=ipos.positions, cell=[0, 0, 0], pbc=False)
                atoms.calc = SinglePointCalculator(atoms, energy=ipos.info['e'])
                if force: atoms.arrays['forces']=ipos.arrays['forces']
                poscarout.extend([atoms])
            ase_write(output_file, poscarout, format="extxyz")
            print("Writing %s" %(output_file))
        except Exception as e:
            print(f"Error in extxyz: {e}")
#------------------------------------------------------------------------------------------



import random
import numpy as np
from ase.data              import atomic_numbers , covalent_radii 
from aegon.libcrossover    import get_gen
from aegon.libutils        import rename, writexyzs, rotate_vector_angle_deg
from aegon.libsel_roulette import get_roulette_wheel_selection
from aegon.libdiscusr      import molin_sim_molref, comparator_usr_serial
#-------------------------------------------------------------------------------
def separa_solapamientos(mvatoms,atoms, min_dist_factor=1.1, max_iter=50,doc='traj'):
    atoms_list_out=[]
    #atoms_list_out.extend([atoms])
    for _ in range(max_iter):
        moved = False
        positions = atoms.get_positions()
        num_atoms = len(atoms)
        for i in mvatoms:
            r1 = covalent_radii[atomic_numbers[atoms[i].symbol]]
            for j in range(num_atoms):
                if j != i:
                  dist_vec = positions[j] - positions[i]
                  r2 = covalent_radii[atomic_numbers[atoms[j].symbol]]
                  dist1 = np.linalg.norm(dist_vec)
                  dist2 = dist1/(r1 + r2)
                  if dist2 < min_dist_factor:
                      direction = dist_vec / dist1
                      move_vec = 0.5 * (min_dist_factor - dist2) * direction
                      positions[i] -= move_vec
                      if j in mvatoms:
                          positions[j] += move_vec
                      moved = True
        atoms.set_positions(positions)
        trial=atoms.copy()                #
        atoms_list_out.extend([trial])    #
        if moved is False:
            break
    writexyzs(atoms_list_out, doc) #
    return(atoms)
#-------------------------------------------------------------------------------
#MUTATION 1: Atoms displacement
def atom_displacement(moleculein, delta=0.4, min_dist_factor=1.1, max_iter=50):
    moleculeout = moleculein.copy()
    natoms = len(moleculeout)
    atoms_index = list(range(natoms))
    random.shuffle(atoms_index)
    num_mut = random.randint(1, natoms)
    mvatoms = atoms_index[:num_mut]
    #print(mvatoms)
    for ii in mvatoms:
        desplazamiento = np.random.uniform(-delta, delta, size=3)
        moleculeout.positions[ii] += desplazamiento
    moleculeout = separa_solapamientos(mvatoms,moleculeout, min_dist_factor, max_iter,doc='Mutation_one.xyz')
    moleculeout.info['o']={'i': moleculein.info['i'], 'f':moleculein.info.get('f',0)}
    return moleculeout
#----------------------------------------------------------------------
#MUTATION 2: twist mutation
def twist_mutation(atoms, delta=0.4, min_dist_factor=1.1, max_iter=50):
    atoms_mut = atoms.copy()
    atoms_mut.translate(-atoms_mut.get_center_of_mass())
    up, dn = get_gen(atoms_mut)
    up_indices=up.info['c']
    angulo = random.randint(5, 355)
    eje = [0, 0, 1]
    lista=[]
    for q in np.linspace(0, angulo, 50):
        upcopy=up.copy()
        up_rot=rotate_vector_angle_deg(upcopy, eje, q)
        atoms_mut=up_rot+dn
        atoms_mut.info['e']=atoms.info['e']
        atoms_mut.info['i']=atoms.info['i']
        lista.append(atoms_mut)
    writexyzs(lista, 'Mutation_two1.xyz') #
    atoms_mut = separa_solapamientos(up_indices, atoms_mut, min_dist_factor, max_iter, doc='Mutation_two2.xyz')
    atoms_mut.info['o']={'i': atoms.info['i'], 'f':atoms.info.get('f',0)}
    return atoms_mut
#------------------------------------------------------------------------------------------
def focused_expansion(molin, index, option='expansion'):
    """
    Vectorized version
    """
    molout = molin.copy()
    positions = molout.positions
    vector = positions[index].copy()
    vectors = positions - vector
    distances = np.linalg.norm(vectors, axis=1, keepdims=True)
    distances = np.where(distances == 0, 1e-10, distances)
    rnorm = vectors / distances
    signo=+1 if option == 'expansion' else -1
    edel = signo*np.exp(-distances) * 1.2
    new_positions = positions + rnorm * edel
    new_positions[index] = positions[index]
    molout.positions = new_positions
    # Genera los traj paara el 2
    return molout
#-------------------------------------------------------------------------
#MUTATION 4: atom permutation
def atom_permutation(moleculein):
    moleculeout = moleculein.copy()
    natoms = len(moleculeout)
    species = {}
    for i in range(natoms):species.setdefault(moleculeout.symbols[i], []).append(i)
    el_a, el_b = random.sample(list(species.keys()), 2)
    grupo_a = species[el_a]
    grupo_b = species[el_b]
    ia = random.choice(grupo_a)
    ib = random.choice(grupo_b)
    moleculeout.symbols[ia], moleculeout.symbols[ib] = moleculeout.symbols[ib], moleculeout.symbols[ia]
    ra = covalent_radii[atomic_numbers[moleculeout[ia].symbol]]
    rb = covalent_radii[atomic_numbers[moleculeout[ib].symbol]]
    opt_a, opt_b = ('expansion', 'impasion') if ra > rb else ('impasion', 'expansion')
    moleculeout = focused_expansion(moleculeout, ia, option=opt_a)
    moleculeout = focused_expansion(moleculeout, ib, option=opt_b)
    moleculeout.info['o']={'i': moleculein.info['i'], 'f':moleculein.info.get('f',0)}
    return moleculeout
#--------------------------------------------------------------------------
def inequivalent_different_finder(moleculein, tolerance=0.3, eigen_tolerance=0.008, matrix_tolerance=0.1):
    moleculeout = moleculein.copy()
    natoms = len(moleculeout)
    list_ineq   = inequivalent_finder(moleculeout, tolerance, eigen_tolerance, matrix_tolerance)
    if len(list_ineq) == 0:
        return(False)
    else:
        species = {}
        for i in list_ineq:species.setdefault(moleculeout.symbols[i], []).append(i)
        return(species)
#-------------------------------------------------------------------------------
def make_mutants(the_chosen_ones_list):
    moleculeout=[]
    for imol in the_chosen_ones_list:
        especies= list(set(imol.symbols))
        wich = random.choice([0,1,2]) if len(especies)>1 else random.choice([0,1])
        if wich == 0:
            muty = atom_displacement(imol, delta=0.4, min_dist_factor=1.1, max_iter=50)
            muty.info['Mutation'] = 'Atom Displacement'
        if wich == 1:
            muty = twist_mutation(imol, delta=0.4, min_dist_factor=1.1, max_iter=50)
            muty.info['Mutation'] = 'Twist Mutation'
        if wich == 2:
            muty = atom_permutation(imol)
            muty.info['Mutation'] = 'Atom Permutation'
        moleculeout.append(muty)
    moleculeout = rename(moleculeout,'pre_mutant_',4)
    return moleculeout
#-------------------------------------------------------------------------------
def popgen_cukoo_mutis(moleculein,memory_pre, memory_opt,index,number_of_mutants,tol_gen):
    if number_of_mutants == 0:
        moleculeout = []
        return moleculeout
    #silence = 1 if verbose == 0 else 0 # ¿Que es verbose?
    print("------------------------------------------------------------------")
    print("-------------------------GENERATOR OF MUTANTS---------------------")
    print("------------------------------------------------------------------")
    the_chosen_ones = get_roulette_wheel_selection(moleculein, number_of_mutants * 10)
    moleculeout = make_mutants(the_chosen_ones)
    #print ("Compare with itself: %d" % (len(moleculeout)))
    moleculeout = comparator_usr_serial(moleculeout, tol_gen, 1)
    #print ("Compare with memory_pre: %d" % (len(moleculeout)))
    moleculeout = molin_sim_molref(moleculeout, memory_pre, tol_gen, 1)
    #print ("Compare with memory_opt: %d" % (len(moleculeout)))
    moleculeout = molin_sim_molref(moleculeout, memory_opt, tol_gen, 1)
    moleculeout = moleculeout[0:number_of_mutants]
    moleculeout = rename(moleculeout, 'mutant_'+str(index).zfill(2),4)
    print ("We have %d molecules type MUTANT from %d solicited." % (len(moleculeout), number_of_mutants))
    if len(moleculeout)!=0:
        for imol in moleculeout:
            print("%s by: %-18s in %15s (f=%3.2f)" % (imol.info['i'],imol.info['Mutation'],imol.info['o']['i'],imol.info['o']['f']))
    return moleculeout
#-----------------------------------------------------------------------
def popgen_cukoo_childs(moleculein, memory_pre, memory_opt,index,number_of_childs,tol_gen,atomlist):
    if number_of_childs==0 or len(moleculein)<2:
        moleculeout=[]
        return moleculeout
    moleculeout=[]
    print("------------------------------------------------------------------")
    print("------------------------ MATING: CROSSOVER -----------------------")
    print("------------------------------------------------------------------")
    print("CONSTRUCTION OF THE PARENT LIST THROUGH ROULETTE")
    list_m=get_roulette_wheel_selection(moleculein, number_of_childs*2)
    list_p=get_roulette_wheel_selection(moleculein, number_of_childs*2)
    for i in range(number_of_childs*2):
        cross = crossover_deavenho(list_p[i], list_m[i], atomlist)
        cross.info['parents']={'p': list_p[i].info['i'], 'm':list_m[i].info['i']}
        cross.info['f_parents']={'p': list_p[i].info['f'], 'm': list_m[i].info['f']}
        if cross:  moleculeout.extend([cross])
    for atoms in moleculeout: atoms.info['e'] = 0
    moleculeout = comparator_usr_serial( moleculeout, tol_gen, 1)
    moleculeout = molin_sim_molref( moleculeout, memory_pre, tol_gen, 1)
    moleculeout = molin_sim_molref( moleculeout, memory_opt, tol_gen, 1)
    moleculeout=moleculeout[0:number_of_childs]
    moleculeout=rename( moleculeout,'mating_'+str(index).zfill(2),4)
    print("We have %d molecules type MATING from %d solicited" %(len(moleculeout), number_of_childs))
    if len(moleculeout)!=0:
        for imol in moleculeout:
            print("%s mating: %15s (f=%3.2f) and %15s (f=%3.2f)" %(imol.info['i'], imol.info['parents']['m'], imol.info['f_parents']['m'], imol.info['parents']['p'], imol.info['f_parents']['p']))
    return(moleculeout)
    # ordenar por energía el sumarry generatirion  ...


import os
import re
import numpy as np 
from ase import Atom, Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from aegon.libutils import sort_by_energy
#------------------------------------------------------------------------------------------
hartree2eV = 27.211386245981 #NIST
bohr2angstrom=0.529177210544 #NIST
eVtokcalpermol=23.060548012069496
hartree2kcalmol=627.5094738898777
#------------------------------------------------------------------------------------------
def get_termination_orca(pathfilename):
    if os.path.isfile(pathfilename):
        opt=0
        freq=0
        file=open(pathfilename,'r')
        for line in file:
            if "OPTIMIZATION RUN DONE" in line: opt=opt+1
            if "VIBRATIONAL FREQUENCIES" in line: freq=freq+1
        file.close()
        total=opt+freq
        return total
    else:
        return -1
#------------------------------------------------------------------------------------------
def get_energy_orca(pathfilename):
    eneinEh=0.0
    file = open(pathfilename,'r')
    for line in file:
        if 'Total Energy' in line:
            ls = line.split()
            if ls[4]=='Eh': eneinEh = float(ls[3])
    file.close()
    return eneinEh
#------------------------------------------------------------------------------------------
def get_geometry_orca(pathfilename):
    nt=get_termination_orca(pathfilename)
    if nt==-1: return False
    filename=os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    ene=get_energy_orca(pathfilename)
    file=open(pathfilename,'r')
    for line in file:
        if "CARTESIAN COORDINATES (ANGSTROEM)" in line:
            singlemol = Atoms()
            singlemol.info['e'] = ene     #ENERGY IN Eh
            singlemol.info['i'] = namein
            singlemol.info['t'] = nt
            line=file.readline()
            line=file.readline()
            ls = line.split()
            while len(ls)>0:
                ss,xc,yc,zc = ls[0],float(ls[1]), float(ls[2]), float(ls[3])
                ai=Atom(symbol=ss,position=(xc, yc, zc))
                singlemol.append(ai)
                line=file.readline()
                ls = line.split()
    file.close()
    return singlemol
#------------------------------------------------------------------------------------------
def get_traj_orca(pathfilename, force=False):
    filename = os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    N = 0
    start, end, ene, start_2, end_2 = [], [], [], [], []
    openold = open(pathfilename,"r")
    rline = openold.readlines()
    for i in range(len(rline)):
        if "GEOMETRY OPTIMIZATION CYCLE" in rline[i]:
            N +=1
        if "CARTESIAN COORDINATES (ANGSTROEM)" in rline[i]:
            start.append(i+2)
            for j in range(i + 2, len(rline)):
                if rline[j].strip() == "":
                    end.append(j - 1)
                    break
        if "TOTAL SCF ENERGY" in rline[i]:
            eneline = rline[i+3].split()
            ene.append(eneline[3])   #ENERGY IN Eh
        if "CARTESIAN GRADIENT" in rline[i] and force:
            start_2.append(i+3)
            for j in range(i + 3, len(rline)):
                if rline[j].strip() == "":
                    end_2.append(j - 1)
                    break
    moleculeout=[]
    for i in range(N):
        singlemol = Atoms()
        singlemol.info['e'] = float(ene[i]) #ENERGY IN Eh
        singlemol.info['i'] = namein+'_'+str(i+1).zfill(3)
        for line in rline[start[i] : end[i]+1]:
            words = line.split()
            ss = str(words[0])
            xc,yc,zc = float(words[1]), float(words[2]), float(words[3])
            ai=Atom(symbol=ss,position=(xc, yc, zc))
            singlemol.append(ai)
        if force:
            forces_list_by_group = []
            for line in rline[start_2[i] : end_2[i]+1]:
                words = line.split()
                fx,fy,fz = float(words[3]), float(words[4]), float(words[5])
                fx=-fx/bohr2angstrom #Eh/A
                fy=-fy/bohr2angstrom #Eh/A
                fz=-fz/bohr2angstrom #Eh/A
                forces_list_by_group.append([fx,fy,fz])
            singlemol.arrays['forces'] = np.array(forces_list_by_group)
        moleculeout.extend([singlemol])
    openold.close()
    return (moleculeout)
#------------------------------------------------------------------------------------------
def get_freqneg_orca(pathfilename):
    with open(pathfilename, 'r') as f:
        contenido = f.read()
    # Search the frequencies after "VIBRATIONAL FREQUENCIES"
    frecuencias = re.findall(r'\d+:\s+(\d+\.\d+)\s+cm\*\*-1', contenido)
    freq_negative=0
    freq_neg_list=[]
    for f in frecuencias:
        if float(f) < 0.0:
            freq_negative=freq_negative+1
            freq_neg_list.append(float(ifreq))
    freq_neg_list.sort()
    freq_sample= freq_neg_list[0] if (freq_negative > 0) else 0.0
    return freq_negative, freq_sample
#------------------------------------------------------------------------------------------
#------------------------------------------------------------------------------------------
#------------------------------------------------------------------------------------------
def read_traj_coords(pathfilename_coordinates):
    file=open(pathfilename_coordinates,'r')
    moleculeout=[]
    unicos=[]
    for line in file:
        ls=line.split()
        if len(ls)==1:
            natoms=int(ls[0])
            count=0
            line=file.readline()
            ls=line.split()
            k=int(ls[5].split(',')[0])
            mol = Atoms()
        if len(ls)==4:
            ss = str(ls[0])
            xc,yc,zc = float(ls[1]), float(ls[2]), float(ls[3])        #Angstrom
            ai=Atom(symbol=ss, position=(xc, yc, zc))
            mol.append(ai)
            count=count+1
            if (count==natoms) and k not in unicos:
                moleculeout.extend([mol])
                unicos.append(k)
    file.close()
    return moleculeout
#------------------------------------------------------------------------------------------
def read_traj_forces(pathfilename_forces):
    file=open(pathfilename_forces,'r')
    allforces=[]
    unicos=[]
    for line in file:
        ls=line.split()
        if len(ls)==1:
            natoms=int(ls[0])
            count=0
            line=file.readline()
            if "Unit is kJ mol^-1 Angstrom^-1" in line:
                hartree2joule=4.359744722206
                avogadro=6.02214076
                factor=1.0/(hartree2joule*avogadro*100.0)
            elif "Unit is Hartree/Angstrom" in line:
                factor=1.0
            else:
                print(pathfilename_forces, "ERROR Units")
            ls=line.split()
            k=int(ls[5].split(',')[0])
            forces_list_by_group = []
        if len(ls)==4:
            ss = str(ls[0])
            fx,fy,fz = float(ls[1]), float(ls[2]), float(ls[3])
            fx=-fx*factor   #FORCE IN Eh/A
            fy=-fy*factor   #FORCE IN Eh/A
            fz=-fz*factor   #FORCE IN Eh/A
            forces_list_by_group.append([fx,fy,fz])
            count=count+1
            if (count==natoms) and k not in unicos:
                allforces.append(forces_list_by_group)
                unicos.append(k)
    file.close()
    return allforces
#------------------------------------------------------------------------------------------
def read_traj_enes(pathfilename_out):
    energy_list = []
    openold_e = open(pathfilename_out,"r")
    for line in openold_e:
        if "Step |  Sim. Time | Iter |  t_Ener |  t_Grad |     Temp |     E_Kin |         E_Pot |         E_Tot |      Cons.Qty | E.Drift" in line:
            line=openold_e.readline()
            line=openold_e.readline()
            line=openold_e.readline()
            while line.startswith(" "):
                ls = line.split()
                ene=float(ls[6]) if ls[0]=='0' else float(ls[7]) #ENERGY IN Eh
                energy_list.append(ene)
                line=openold_e.readline()
    return energy_list
#------------------------------------------------------------------------------------------
def get_extxyz(pathfilename_coordinates, pathfilename_forces, pathfilename_out):
    ene=read_traj_enes(pathfilename_out)
    frc=read_traj_forces(pathfilename_forces)
    xyz=read_traj_coords(pathfilename_coordinates)
    #filename_one = os.path.basename(pathfilename_out)
    #namein_one = filename_one.split('.')[0]
    atoms_list_out=[]
    if len(ene) == len(frc) == len(xyz):
        for i in range(len(ene)):
            imol=xyz[i]
            atoms = Atoms(imol.symbols, positions=imol.positions, cell=[0, 0, 0], pbc=False)
            atoms.calc = SinglePointCalculator(atoms, energy=ene[i])
            atoms.arrays['forces']=np.array(frc[i])
            #atoms.info['i']=namein_one+'_'+str(i+1).zfill(3)
            atoms_list_out.extend([atoms])
    else: print('length error in %s' %(pathfilename_out))
    return atoms_list_out
#------------------------------------------------------------------------------------------
def example():
    import glob
    from ase.io import read, write
    files=sorted(glob.glob('*.out'))
    atoms_list_out=[]
    for ifile in files:
        basename=ifile.split('.')[0]
        traj=get_extxyz(basename+'.traj.xyz', basename+'.frc.xyz', basename+'.out')
        write(basename+'_ext.xyz', traj, format="extxyz")
        atoms_list_out=atoms_list_out+traj
    write('traj_ext.xyz', atoms_list_out, format="extxyz")
#example()



import os
import time
import numpy as np
from ase import Atoms
from multiprocessing import Process
from aegon.libutils import sort_by_energy, prepare_folders, split_poscarlist
#------------------------------------------------------------------------------------------
def make_ainput_gulp(headgulp, singleposcar, folder):
    iposname=singleposcar.info['i']
    nameinp=iposname+'.gin'
    nameout=iposname+'.got'
    if os.path.isfile(folder+nameout):
        print ("%s ... was not built because %s exists" %(folder+nameinp, folder+nameout))
    elif os.path.isfile(folder+nameinp):
        print ("%s .... was not built because it exists." %(nameinp))
    else:
        #print("Making input file = %s" %(folder+nameinp))
        fh=open(folder+nameinp,"w")
        for iline in headgulp:
            if iline=='LATTICEVECTORS':
                matrix=singleposcar.cell
                print("%12.9f %16.9f %16.9f" %(matrix[0,0],matrix[0,1],matrix[0,2]), file=fh)
                print("%12.9f %16.9f %16.9f" %(matrix[1,0],matrix[1,1],matrix[1,2]), file=fh)
                print("%12.9f %16.9f %16.9f" %(matrix[2,0],matrix[2,1],matrix[2,2]), file=fh)
            elif iline=='COORDINATES':
                symbols = singleposcar.get_chemical_symbols()
                direct_coordinates = singleposcar.get_scaled_positions()
                for symbol, coord in zip(symbols, direct_coordinates):
                    print("%-2s %16.9f %16.9f %16.9f" %(symbol,coord[0], coord[1], coord[2]), file=fh)
            else:
                print(iline, file=fh)
        fh.close()
#------------------------------------------------------------------------------------------
def make_the_sh_gulp(basename, poscarlist, gulp_path):
    filesh=basename+'.sh'
    if os.path.isfile(filesh):
        os.system("rm -f -v "+filesh)
    else:
    #if not os.path.isdir(basename):
        fh=open(filesh,'w')
        print("#!/bin/bash", file=fh)
        print("exe_gulp=%s" %(gulp_path), file=fh)
        print("cd %s" %(basename), file=fh)
        for iposcar in poscarlist:
            print("wait $PID ; $exe_gulp < %s.gin > %s.got" %(iposcar.info['i'],iposcar.info['i']), file=fh)
        print("cd ..", file=fh)
        print("exit 0", file=fh)
        fh.close()
#------------------------------------------------------------------------------------------
def get_termination_gulp(pathfilename):
    if os.path.isfile(pathfilename):
        file=open(pathfilename,'r')
        for line in file:
            if '**** Optimisation achieved ****' in line:
                normal = 1
                break
            else:
                normal = 0
        file.close()
        return normal
    else:
        return False
#------------------------------------------------------------------------------------------
def get_energy_gulp(pathfilename):
    file = open(pathfilename,'r')
    for line in file:
        if 'Final energy =' in line:
            ls = line.split()
            eneineV = float(ls[3])
        if 'Final enthalpy =' in line:
            ls = line.split()
            eneineV = float(ls[3])
    file.close()
    #eneinkcalpermol=eneineV*eVtokcalpermol
    #return eneinkcalpermol
    return eneineV
#------------------------------------------------------------------------------------------
def get_geometry_gulp(pathfilename):
    nt=get_termination_gulp(pathfilename)
    if nt==False: return False
    filename = os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    matrix=np.array([])
    file = open(pathfilename,'r')
    ans=False
    ans1 = 'Final Cartesian lattice vectors (Angstroms) :'
    ans2 = 'Cartesian lattice vectors (Angstroms) :'
    for line in file:
        if (ans1 in line) or (ans2 in line):
            ans=True
            line=file.readline()
            line=file.readline()
            ls = line.split()
            a1x, a1y, a1z=float(ls[0]), float(ls[1]), float(ls[2])
            line=file.readline()
            ls = line.split()
            a2x, a2y, a2z=float(ls[0]), float(ls[1]), float(ls[2])
            line=file.readline()
            ls = line.split()
            a3x, a3y, a3z=float(ls[0]), float(ls[1]), float(ls[2])
            matrix=np.array([[a1x, a1y, a1z],[a2x, a2y, a2z],[a3x, a3y, a3z]])
    file.close()
    file=open(pathfilename,'r')
    direct_coordinates,symbols=[],[]
    for line in file:
        if 'Final fractional coordinates of atoms :' in line:
            for ii in range(5): line=file.readline()
            line=file.readline()
            ls = line.split()
            lenls=len(ls)
            while ( lenls == 7 ):
                si=str(ls[1])
                symbols.append(si)
                xd,yd,zd=float(ls[3]),float(ls[4]),float(ls[5])
                direct_coordinates.append([xd,yd,zd])
                line=file.readline()
                ls = line.split()
                lenls=len(ls)
    file.close()
    poscarx = Atoms(symbols=symbols, scaled_positions=direct_coordinates, cell=matrix, pbc=ans)
    energy=get_energy_gulp(pathfilename)
    poscarx.info['c'] = nt
    poscarx.info['i'] = namein
    poscarx.info['e'] = energy
    return poscarx
#------------------------------------------------------------------------------------------
def get_all_xt_geometry_gulp(poscarlist, path):
    poscarout,count=[],0
    for ipos in poscarlist:
        iposname=ipos.info['i']
        filename = iposname + '.got'
        pos01 = get_geometry_gulp(path+filename)
        if pos01:
            poscarout.extend([pos01])
            count = count + 1
    if count==0: return False
    else: return poscarout
#------------------------------------------------------------------------------------------
def structure_recolector(folders, poscars_split):
    list_ne, list_at, list_nt, listall=[],[],[],[]
    ntot=0
    moleculeout=[]
    for ifolder, iposcars in zip(folders,poscars_split):
        ntot=ntot+len(iposcars)
        molx=get_all_xt_geometry_gulp(iposcars, ifolder+'/')
        if molx is not False:
            moleculeout=moleculeout+molx
        for iposcar in iposcars:
            id = get_termination_gulp(ifolder+'/'+iposcar.info['i']+'.got')
            if (id is False): list_ne.append(iposcar.info['i'])
            elif (id == 0):   list_at.append(iposcar.info['i'])
            elif (id == 1):   list_nt.append(iposcar.info['i'])
    nnt, nat, nne = len(list_nt), len(list_at), len(list_ne)
    pnt=int(float(nnt)*100.0/float(ntot))
    pat=int(float(nat)*100.0/float(ntot))
    print("Calculations with N.T. = %d (%d percent)" %(nnt,pnt))
    print("Calculations with A.T. = %d (%d percent)" %(nat,pat))
    if len(moleculeout)==0:
        print("ZERO calculation finished satisfactorily")
        exit()
    return moleculeout
#------------------------------------------------------------------------------------------
def send_sh_file(sh_file):
    os.system("bash %s" %(sh_file))
    os.system("rm -f %s" %(sh_file))
    return 0
#------------------------------------------------------------------------------------------
def calculator_all_gulp(poscar_list_in, block_gulp, gulp_path, nproc, base_name):
    start = time.time()
    folder_names=prepare_folders(poscar_list_in, nproc, base_name)
    split_poscars= split_poscarlist(poscar_list_in, nproc)
    for poscars, ifolder in zip(split_poscars, folder_names):
        for iposcar in poscars:
            make_ainput_gulp(block_gulp, iposcar, ifolder+'/')
        make_the_sh_gulp(ifolder, poscars, gulp_path)
    procs = []
    print("Local submit: %sproc01.sh to %sproc%s.sh" %(base_name, base_name, str(nproc).zfill(2)), end=" ")
    for ifolder in folder_names:
        if os.path.isfile(ifolder+'.sh'):
            proc = Process(target=send_sh_file, args=(ifolder+'.sh',))
            procs.append(proc)
            proc.start()
    for proc in procs:
        proc.join()
    end = time.time()
    print('TOTAL TIME: %5.2f s' %(end - start))
    poscar_list_out=structure_recolector(folder_names, split_poscars)
    return poscar_list_out
#------------------------------------------------------------------------------------------
def display_info(moleculein, stage_string, dicc_term):
    print("-------------------------------------------------------------------")
    print("------------------------- SUMMARY %s -------------------------" %(stage_string))
    print("Number File--------Name   Energy (ev)   Delta----E T")
    molzz = sort_by_energy(moleculein, 1)
    emin = molzz[0].info['e']
    for ii, imol in enumerate(molzz):
        ei = imol.info['e']
        id = imol.info['i']
        #nt = imol.info['c']
        nt = dicc_term[id]
        deltae  =  ei - emin
        kk=str(ii+1).zfill(6)
        print("%s %s %13.8f %12.8f %d" %(kk, id, ei, deltae, nt))
#------------------------------------------------------------------------------------------------



import os
import time
import numpy as np
from multiprocessing import Process
from dscribe.descriptors import MBTR, ValleOganov
from aegon.libposcar import readposcars, writeposcars
from aegon.libutils import prepare_folders, split_poscarlist
#------------------------------------------------------------------------------------------
def find_similar_elements(similarity_matrix, threshold):
    similar_elements_indices = []
    num_elements = similarity_matrix.shape[0]
    for i in range(num_elements):
        for j in range(i+1,num_elements):
            if similarity_matrix[i, j] >= threshold:
                similar_elements_indices.append(j)
    return similar_elements_indices
#------------------------------------------------------------------------------------------
def disc_MBTR(atoms, threshold, nproc=1):
    num_molecules = len(atoms)
    num_atomsxmol = len(atoms[0])
    species = set(atoms[0].get_chemical_symbols())
    r_cut=10
    geometry={"function": "distance"}
    grid={"min": 0, "max": r_cut, "sigma": 1E-5, "n" : 100}
    weighting = {"function": "inverse_square", "r_cut": r_cut,  "threshold": 1E-3}
    opt="none"
    #mbtr=MBTR(species=species, geometry=geometry, weighting=weighting,  grid=grid, periodic=True, normalization=opt, normalize_gaussians=True, sparse=False, dtype="float32")
    mbtr=ValleOganov(species=species, function='distance', n=100, sigma=1E-5, r_cut=r_cut)
    n_features=mbtr.get_number_of_features()
    if nproc ==1:
        descriptors=[mbtr.create(imol) for imol in atoms]
    elif nproc > 1:
        descriptors=mbtr.create(atoms, n_jobs=nproc)
    similar_elements_indices = []
    similarity_matrix = np.zeros((num_molecules, num_molecules))
    for i in range(num_molecules):
        vi=descriptors[i]
        for j in range(i, num_molecules):
            vj=descriptors[j]
            manhattan_distance=sum([np.absolute(a-b) for a,b in zip(vi,vj)])
            similarity=1.0/(1.0+manhattan_distance/float(n_features))
            similarity_matrix[i, j] = similarity
            similarity_matrix[j, i] = similarity
    similar_elements_indices=find_similar_elements(similarity_matrix, threshold)
    similar_elements_indices.sort()
    #print(similar_elements_indices)
    disimilars_atoms=[atoms[i] for i in range(num_molecules) if i not in similar_elements_indices]
    return disimilars_atoms
#------------------------------------------------------------------------------------------
def comparator_mbtr_conv(atoms0, threshold, nproc):
    start = time.time()
    ni=len(atoms0)
    atoms1=disc_MBTR(atoms0, threshold, nproc)
    nf=len(atoms1)
    end = time.time()
    print('Comparator MBTR conv at %5.2f s [%d -> %d]' %(end - start, ni, nf))
    return atoms1
#------------------------------------------------------------------------------------------
def make_comparator_mbtr(ifolder, threshold):
    atoms0=readposcars(ifolder+'/'+ifolder+'.vasp')
    atoms1=disc_MBTR(atoms0, threshold, 1)
    writeposcars(atoms1,ifolder+'/'+ifolder+'_disc.vasp', 'D', 1)
#------------------------------------------------------------------------------------------
def comparator_mbtr_parallel(poscarlist, threshold, nproc, base_name):
    start = time.time()
    ni=len(poscarlist)
    folderlist=prepare_folders(poscarlist, nproc, base_name)
    poscar_split_list=split_poscarlist(poscarlist, nproc)
    procs = []
    #dicc_term = {iposcar.info['i']: iposcar.info['c'] for iposcar in poscarlist}
    for ifolder, iposcars in zip(folderlist, poscar_split_list):
        writeposcars(iposcars, ifolder+'/'+ifolder+'.vasp', 'D', 1)
        proc = Process(target=make_comparator_mbtr, args=(ifolder,threshold,))
        procs.append(proc)
        proc.start()
    for proc in procs:
        proc.join()
    moleculeout=[]
    for ifolder in folderlist:
        molx=readposcars(ifolder+'/'+ifolder+'_disc.vasp')
        #for imol in molx: imol.info['c']=dicc_term[imol.info['i']]
        moleculeout=moleculeout+molx
    #os.system('rm -rf %sproc[0-9][0-9]' %(base_name))
    nf=len(moleculeout)
    end = time.time()
    print('MBTR comparison (parallel) at %5.2f s [%d -> %d]' %(end - start, ni, nf))
    return moleculeout
#------------------------------------------------------------------------------------------
def make_similarity_matrix_compare(moleculein, moleculeref, nproc=1):
    num_molecules = len(moleculein)
    num_atomsxmol = len(moleculein[0])
    species = set(moleculein[0].get_chemical_symbols())
    r_cut, sigma, n=10, 1E-5, 100
    geometry={"function": "distance"}
    grid={"min": 0, "max": r_cut, "sigma": sigma, "n" : n}
    weighting = {"function": "inverse_square", "r_cut": r_cut,  "threshold": 1E-3}
    opt="none"
    #mbtr=MBTR(species=species, geometry=geometry, weighting=weighting,  grid=grid, periodic=True, normalization=opt, normalize_gaussians=True, sparse=False, dtype="float32")
    mbtr=ValleOganov(species=species, function='distance', n=n, sigma=sigma, r_cut=r_cut)
    n_features=mbtr.get_number_of_features()
    if nproc ==1:
        descriptors1=[mbtr.create(imol) for imol in moleculein]
        descriptors2=[mbtr.create(imol) for imol in moleculeref]
    elif nproc > 1:
        descriptors1=mbtr.create(moleculein, n_jobs=nproc)
        descriptors2=mbtr.create(moleculeref, n_jobs=nproc)
    total_molecules1=len(moleculein)
    total_molecules2=len(moleculeref)
    similarity_matrix=np.zeros(shape=(total_molecules1, total_molecules2),dtype=float)
    for i in range(total_molecules1):
        vi=descriptors1[i]
        for j in range(total_molecules2):
            vj=descriptors2[j]
            manhattan_distance=sum([np.absolute(a-b) for a,b in zip(vi,vj)])
            similarity=1.0/(1.0+manhattan_distance/float(n_features))
            similarity_matrix[i, j] = similarity
    return similarity_matrix
#------------------------------------------------------------------------------------------
def molin_sim_molref(moleculein, moleculeref, tols=tolsij, tole=tolene, nproc=1):
    start = time.time()
    ni=len(moleculein)
    matrixs=make_similarity_matrix_compare(moleculein, moleculeref, nproc)
    similares=[]
    for i,imol in enumerate(moleculein):
        for j, jmol in enumerate(moleculeref):
            sij=matrixs[i][j]
            edf=np.abs(imol.info['e']-jmol.info['e'])
            if (sij >= tols) and (edf <= tole):
                similares.append(i)
                break
    moleculeout=[imol for i, imol in enumerate(moleculein) if i not in similares] 
    nf=len(moleculeout)
    end = time.time()
    print('MBTR comparison (-serial-) at %5.2f s [%d -> %d]' %(end - start, ni, nf))
    return moleculeout
#------------------------------------------------------------------------------------------
